/*******************************************************************
 *
 *	File:		Video.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <math.h>
#include <stdlib.h>
#include <coedef.h>		// for ECoeWinPriorityNeverAtFront
//#include <QuartzKeys.h>	// MR
#include <gdi.h>
#include "video.h"
#include "gameengine.h"
#include "navigation.h"

_LIT(KFontName,"Courier");
const TInt KOneSecondInMicroSeconds =	1000 * 1000;

#ifndef __WINS__
extern "C" void DEB_Print(int i, int j, int k, int l);
#endif	// __WINS__

/*
 * TODO:
 * - orientation
 * - vk symbols
 * - vk buttons
 * - add a MGameTickObserver. Some of the MGameScreenDrawObserver should be changed to be 
 * a MGameTickObserver instead
 * - use proper optimisation options
 *		 -O3 -fstrength-reduce -fthread-jumps  -fcse-follow-jumps -fcse-skip-blocks -frerun-cse-after-loop  -fschedule-insns -fschedule-insns2 -fexpensive-optimizations  -fforce-mem -fforce-addr -funroll-loops
 */


/*****************************************
 *
 * TScaledBitmapContext implementation
 *
 *****************************************/

TScaledBitmapContext::TScaledBitmapContext() : iSkipX(-1), iSkipY(-1), iSrcGameScreenDevice(NULL)
{
}


TScaledBitmapContext::TScaledBitmapContext(const CGameScreenDevice& aDstGameScreenDevice, const CGameScreenDevice& aSrcGameScreenDevice, const TSize& aBitmapSize, TBool aCenter)
	: iSrcGameScreenDevice(&aSrcGameScreenDevice)
{
	float	scaleX = ((float) aDstGameScreenDevice.Size().iWidth) / ((float) aBitmapSize.iWidth);
	float	scaleY = ((float) aDstGameScreenDevice.Size().iHeight) / ((float) aBitmapSize.iHeight);
	if ((scaleX < 1) || (scaleY < 1))
	{
		float scale = (scaleX < scaleY) ? scaleX : scaleY;

		// ideal size
		iSize.iWidth = (TInt) (scale * aBitmapSize.iWidth);
		iSize.iHeight = (TInt) (scale * aBitmapSize.iHeight);

		iSkipX = (TInt) floor (((float) aBitmapSize.iWidth) / (((float) aBitmapSize.iWidth) - ((float)iSize.iWidth)));
		iSkipY = (TInt) floor (((float) aBitmapSize.iHeight) / (((float) aBitmapSize.iHeight) - ((float) iSize.iHeight)));

		// real size based on the pixels we are going to skip
		iSize.iWidth = aBitmapSize.iWidth - (aBitmapSize.iWidth / iSkipX);
		iSize.iHeight = aBitmapSize.iHeight - (aBitmapSize.iHeight / iSkipY);

		iSkipY--;	// we start counting at zero
		iSkipX--;
	}

	if (aCenter)
	{
		iPoint.iX = (aDstGameScreenDevice.Size().iWidth - iSize.iWidth) / 2;
		iPoint.iY = (aDstGameScreenDevice.Size().iHeight - iSize.iHeight) / 2;
	}
	iScanLineBytes = iSize.iWidth * (TGameBitmapUtil::GetBpp(aDstGameScreenDevice.DisplayMode())/8);
	ASSERT(iScanLineBytes < (1024 * 4));
}



/*****************************************
 *
 * TGameBitmapUtil implementation
 *
 *****************************************/

TRect TGameBitmapUtil::VisibleIntersection(const TSize& aScreenSize, const TPoint& aOrigin, const TSize& aBitmapSize, const TRect& aBitmapRect)
{
	TRect	screenRect = TRect(aScreenSize);
	TRect	bitmapRect = TRect(aBitmapSize);
	bitmapRect.Intersection(aBitmapRect);
	bitmapRect.Move(aOrigin);
	screenRect.Intersection(bitmapRect);
	return screenRect;
}


TUint8* TGameBitmapUtil::PixelAddress(const CGameScreenDevice* aSrc, const TPoint& aPoint)
{
	return aSrc->ScanLineAddress(aPoint.iY) + ((aPoint.iX * GetBpp(aSrc->DisplayMode())) / 8);
}

TDisplayMode TGameBitmapUtil::GetDisplayMode(TInt aBpp, TBool aColor)
{
	if (aBpp == 1)
		return EGray2;
	else if (aBpp == 2)
		return EGray4;
	else if (aBpp == 4)
		return aColor ? EColor16 : EGray16;
	else if (aBpp == 8)
		return aColor ? EColor256 : EGray256;
	else if (aBpp == 16)
		return EColor64K;
	else if (aBpp == 12)
		return EColor4K;
	else if (aBpp == 24)
		return EColor16M;
	else if (aBpp == 32)
		return ERgb;

	ASSERT(EFalse);
	return ENone;
}

/*
 * GetBpp returns the number of bits we need per pixel in the given
 * display mode. It is different from GetColorDepth in that GetBpp can be
 * used to calculate byte offsets in a bitmap/framebuffer.
 * For example, GetBpp(EColor4K) returns 16 because each pixel requires
 * 16 bits (of which only the least significant 12 bits are used).
 * GetColorDepth(EColor4K) returns 12.
 */
TInt TGameBitmapUtil::GetBpp(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			32,
			32,
			16
		};
	return bpp[aDisplayMode];
}

TInt TGameBitmapUtil::GetColorDepth(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			24,
			32,
			12
		};
	return bpp[aDisplayMode];
}



TUint TGameBitmapUtil::GetColorValue(const TRgb& aColor, const TDisplayMode& aDisplayMode)
{
	switch (aDisplayMode)
	{
		case EColor256:
			return aColor.Color256();
			break;
		case EColor4K:
			return aColor.Color4K();
			break;
		case EColor16M:
			return aColor.Color16M();
			break;
		case EColor64K:
			return aColor.Color64K();
			break;
		case ERgb:
			return aColor.Value();
			break;
		default:
			ASSERT(EFalse);
	}
	return 0;
}


/*****************************************
 *
 * CGamePalette implementation
 *
 *****************************************/

CGamePalette::CGamePalette(TDisplayMode aDisplayMode) : iDisplayMode(aDisplayMode)
{
}

CGamePalette::~CGamePalette()
{
}

CGamePalette* CGamePalette::NewL(TDisplayMode aDisplayMode)
{
	CGamePalette*	self = new(ELeave)CGamePalette(aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}


void CGamePalette::ConstructL()
{
	iPaletteMaxIndex = 256;
	Mem::FillZ(iPalette, sizeof(iPalette));
	iPaletteLastIndexUsed = 0;
}



TInt CGamePalette::AllocateEntry(const TRgb& aColor)
{
	TUint	colorValue = TGameBitmapUtil::GetColorValue(aColor, iDisplayMode);
	for (TInt i=0 ; i<iPaletteLastIndexUsed; i++)
		if (iPalette[i] == colorValue)
			return i;
	ASSERT(iPaletteLastIndexUsed<iPaletteMaxIndex);
	iPalette[iPaletteLastIndexUsed] = colorValue;
	return (iPaletteLastIndexUsed++);
}



/*****************************************
 *
 * CGameScreenDevice implementation
 *
 *****************************************/

void CGameScreenDevice::DirectDraw_16bpp_16bppRotLeft(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize)
{
	TInt					width = aSize.iWidth;
	TInt					height = aSize.iHeight;

	register TUint16*		dstLine = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	register TUint16*		srcLine = (TUint16*) TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	for(int i=0 ; i<height ; i++)
	{
		TUint16*			dst = dstLine;
		TUint16*			src = srcLine;
		for(int j=0 ; j<width ; j++)
		{
			*dst = *src++;
			dst += BytesPerLine()>>1;
		}
		dstLine--;
		srcLine += aSrcDevice.BytesPerLine()>>1;
	}
}

void CGameScreenDevice::DirectDraw_16bpp_16bppRotRight(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize)
{
	TInt					width = aSize.iWidth;
	TInt					height = aSize.iHeight;

	register TUint16*		dstLine = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	register TUint16*		srcLine = (TUint16*) TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	for(int i=0 ; i<height ; i++)
	{
		TUint16*			dst = dstLine;
		TUint16*			src = srcLine;
		for(int j=0 ; j<width ; j++)
		{
			*dst = *src++;
			dst -= BytesPerLine()>>1;
		}
		dstLine++;
		srcLine += aSrcDevice.BytesPerLine()>>1;
	}
}


CGameScreenDevice::CGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode) : iSize(aSize), iDisplayMode(aDisplayMode)
{
}

CGameScreenDevice::~CGameScreenDevice()
{
}


const CGameScreenDevice::TDirectDrawer CGameScreenDevice::KDirectDrawers12or16bpp[KVideoModes] = 
{
	&CGameScreenDevice::DirectDraw_Unsupported,			/* ENone */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* EGray2 */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* EGray4 */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* EGray16 */
	&CGameScreenDevice::DirectDraw_12or16bpp_8bpp,		/* EGray256 */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* EColor16 */
	&CGameScreenDevice::DirectDraw_12or16bpp_8bpp,		/* EColor256 */
	&CGameScreenDevice::DirectDraw_12or16bpp_12or16bpp,	/* EColor64K */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* EColor16M */
	&CGameScreenDevice::DirectDraw_Unsupported,			/* ERgb */
	&CGameScreenDevice::DirectDraw_12or16bpp_12or16bpp	/* EColor4K */
};

const CGameScreenDevice::TDirectDrawer CGameScreenDevice::KDirectDrawers8bpp[KVideoModes] = 
{
	&CGameScreenDevice::DirectDraw_Unsupported,	/* ENone */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EGray2 */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EGray4 */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EGray16 */
	&CGameScreenDevice::DirectDraw_8bpp_8bpp,	/* EGray256 */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EColor16 */
	&CGameScreenDevice::DirectDraw_8bpp_8bpp,	/* EColor256 */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EColor64K */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* EColor16M */
	&CGameScreenDevice::DirectDraw_Unsupported,	/* ERgb */
	&CGameScreenDevice::DirectDraw_Unsupported	/* EColor4K */
};


void CGameScreenDevice::Clear()
{
	TInt	height = iSize.iHeight;
	for (TInt h=0 ; h<height ; h++)
	{
		TUint8*		line = ScanLineAddress(h);
		Mem::FillZ(line, BytesPerLine());
	}
}


void CGameScreenDevice::DirectDraw(const TPoint& aPoint, const CGameScreenDevice& aSrcDevice, const TRect& aRect)
{
	TRect	rect = TGameBitmapUtil::VisibleIntersection(Size(), aPoint, aSrcDevice.Size(), aRect);
	TPoint	dstPoint = aRect.iTl;
	if (aPoint.iX < 0)
		dstPoint.iX -= aPoint.iX;
	if (aPoint.iY < 0)
		dstPoint.iY -= aPoint.iY;

	switch (DisplayMode())
	{
		case EColor256:
		case EGray256:
			(this->*KDirectDrawers8bpp[aSrcDevice.DisplayMode()])(rect.iTl, aSrcDevice, dstPoint, rect.Size());
			break;
		case EColor4K:
		case EColor64K:
			(this->*KDirectDrawers12or16bpp[aSrcDevice.DisplayMode()])(rect.iTl, aSrcDevice, dstPoint, rect.Size());
			break;
		default:
			ASSERT(EFalse);
			break;
	}
}



const CGameScreenDevice::TAlphaBlender CGameScreenDevice::KAlphaBlenders12bpp[KVideoModes] = 
{
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* ENone */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray2 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray4 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray16 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray256 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor16 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor256 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor64K */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor16M */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* ERgb */
	&CGameScreenDevice::AlphaBlend_12bpp_12bpp	/* EColor4K */
};

const CGameScreenDevice::TAlphaBlender CGameScreenDevice::KAlphaBlenders16bpp[KVideoModes] = 
{
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* ENone */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray2 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray4 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray16 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EGray256 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor16 */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor256 */
	&CGameScreenDevice::AlphaBlend_16bpp_16bpp,	/* EColor64K */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* EColor16M */
	&CGameScreenDevice::AlphaBlend_Unsupported,	/* ERgb */
	&CGameScreenDevice::AlphaBlend_Unsupported	/* EColor4K */
};

void CGameScreenDevice::AlphaBlend(const TPoint& aPoint, const CGameScreenDevice& aSrcDevice, const TRect& aRect, TInt8 aAlpha)
{
	TRect	rect = TGameBitmapUtil::VisibleIntersection(Size(), aPoint, aSrcDevice.Size(), aRect);
	TPoint	dstPoint = aRect.iTl;
	if (aPoint.iX < 0)
		dstPoint.iX -= aPoint.iX;
	if (aPoint.iY < 0)
		dstPoint.iY -= aPoint.iY;

	switch (DisplayMode())
	{
		case EColor4K:
			(this->*KAlphaBlenders12bpp[aSrcDevice.DisplayMode()])(rect.iTl, aSrcDevice, dstPoint, rect.Size(), aAlpha);
			break;
		case EColor64K:
			(this->*KAlphaBlenders16bpp[aSrcDevice.DisplayMode()])(rect.iTl, aSrcDevice, dstPoint, rect.Size(), aAlpha);
			break;
		default:
			ASSERT(EFalse);
			break;
	}
}


void CGameScreenDevice::DirectDoubleDraw(const TPoint& /*aPoint*/, const CGameScreenDevice& /*aSrcDevice*/, const TRect& /*aRect*/)
{
#if 0
	TRect	rect = TGameBitmapUtil::VisibleIntersection(Size(), aPoint, aSrcDevice.Size(), aRect);
	TPoint	dstPoint = aRect.iTl;
	if (aPoint.iX < 0)
		dstPoint.iX -= aPoint.iX;
	if (aPoint.iY < 0)
		dstPoint.iY -= aPoint.iY;
#endif
}


const CGameScreenDevice::TScaledScanLineDrawer CGameScreenDevice::KScaledScanLineDrawers12or16bpp[KVideoModes] = 
{
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* ENone */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EGray2 */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EGray4 */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EGray16 */
	&CGameScreenDevice::ScaledScanLineDraw_12or16bpp_8bpp,	/* EGray256 */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EColor16 */
	&CGameScreenDevice::ScaledScanLineDraw_12or16bpp_8bpp,	/* EColor256 */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EColor64K */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* EColor16M */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported,		/* ERgb */
	&CGameScreenDevice::ScaledScanLineDraw_Unsupported		/* EColor4K */
};

void my_memcpy(void* out, const void* in, size_t length)
{
	TUint16*	dst = (TUint16*) out;
	TUint16*	src = (TUint16*) in;

	length>>=1;

	while (length--)
	{
		*dst++ = *src++;
	}
}

void CGameScreenDevice::ScaledDrawBitmap(const TScaledBitmapContext& aScaledBitmapContext)
{
	CGameScreenDevice::TScaledScanLineDrawer scaledScanLineDrawer= NULL;

	switch (DisplayMode())
	{
		case EColor4K:
		case EColor64K:
			scaledScanLineDrawer = KScaledScanLineDrawers12or16bpp[aScaledBitmapContext.SrcGameScreenDevice().DisplayMode()];
			break;
		default:
			break;
	}
	if (!scaledScanLineDrawer)
	{
		ASSERT(EFalse);
		return;
	}


	TUint8*			dst_line = (TUint8*) TGameBitmapUtil::PixelAddress(this, aScaledBitmapContext.Point());
	const CGameScreenDevice& srcGameScreenDevice = aScaledBitmapContext.SrcGameScreenDevice();

	TInt	height = aScaledBitmapContext.SrcGameScreenDevice().Size().iHeight;
	TInt	lineCount = 0;
	TInt	temp=0;

	for (TInt h=0 ; h<height ; h++)
	{
		if (lineCount == aScaledBitmapContext.SkipY())
		{
			(this->*scaledScanLineDrawer)(aScaledBitmapContext, (TUint*)srcGameScreenDevice.ScanLineAddress(h), ETrue);
			lineCount = 0;
		}
		else
		{
			(this->*scaledScanLineDrawer)(aScaledBitmapContext, (TUint*) srcGameScreenDevice.ScanLineAddress(h), EFalse);
			lineCount++;
		}
		if (lineCount != aScaledBitmapContext.SkipY())
		{
			my_memcpy(dst_line, aScaledBitmapContext.ScanLine(), aScaledBitmapContext.ScanLineBytes());
			dst_line += BytesPerLine();
			temp++;
		}
	}
}


void CGameScreenDevice::DirectDraw_12or16bpp_8bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize)
{
	ASSERT(aSrcDevice.Palette());
	ASSERT(iDisplayMode == aSrcDevice.Palette()->DisplayMode());

	TUint*			pal = aSrcDevice.Palette()->PaletteBuffer();
	TUint16*		dst_line = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	const TUint8*	src_line = TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	if (!pal)
		return;

	TInt	height = aSize.iHeight;
	TInt	width = aSize.iWidth;

	for(TInt h=0 ; h<height ; h++)
	{
		TUint16*		dst = dst_line;
		const TUint8*	src = src_line;
		for (TInt w=0 ; w<width ; w++)
		{
			*dst++ = (TUint16) pal[*src++];
		}
		dst_line += (BytesPerLine() >> 1);
		src_line += aSrcDevice.BytesPerLine();
	}
}


void CGameScreenDevice::DirectDraw_8bpp_8bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize)
{
	ASSERT(aSrcDevice.Palette());
	ASSERT(iDisplayMode == aSrcDevice.Palette()->DisplayMode());

	TUint8*			dst_line = (TUint8*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	const TUint8*	src_line = TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	TInt	height = aSize.iHeight;
	TInt	width = aSize.iWidth;

	for(TInt h=0 ; h<height ; h++)
	{
		Mem::Copy(dst_line, src_line, width);
		dst_line += (BytesPerLine() >> 1);
		src_line += aSrcDevice.BytesPerLine();
	}
}




void CGameScreenDevice::DirectDraw_12or16bpp_12or16bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize)
{
	TUint16*		dst_line = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	const TUint16*	src_line = (TUint16*) TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	TInt	height = aSize.iHeight;
	TInt	width = aSize.iWidth;

	for(TInt h=0 ; h<height ; h++)
	{
		TUint16*		dst = dst_line;
		const TUint16*	src = src_line;
		for (TInt w=0 ; w<width ; w++)
		{
			*dst++ = *src++;
		}
		dst_line += (BytesPerLine() >> 1);
		src_line += (aSrcDevice.BytesPerLine() >> 1);
	}
}


void CGameScreenDevice::AlphaBlend_16bpp_16bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize, TUint8 aAlpha)
{
	TUint16*		dst_line = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	const TUint16*	src_line = (TUint16*) TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	TInt	height = aSize.iHeight;
	TInt	width = aSize.iWidth;

	for(TInt h=0 ; h<height ; h++)
	{
		TUint16*		dst = dst_line;
		const TUint16*	src = src_line;
		for (TInt w=0 ; w<width ; w++)
		{
			// alpha blending algorithm "borrowed" from MicroWindows ;-)
			unsigned int s = *src++;
			unsigned int d = *dst;
			unsigned int t = d & 0xf800;
			unsigned int m1, m2, m3;
			m1 = (((((s & 0xf800) - t)*aAlpha)>>8) & 0xf800) + t;
			t = d & 0x07e0;
			m2 = (((((s & 0x07e0) - t)*aAlpha)>>8) & 0x07e0) + t;
			t = d & 0x001f;
			m3 = (((((s & 0x001f) - t)*aAlpha)>>8) & 0x001f) + t;
			*dst++ = (TUint16) (m1 | m2 | m3);
		}
		dst_line += (BytesPerLine() >> 1);
		src_line += (aSrcDevice.BytesPerLine() >> 1);
	}
}

void CGameScreenDevice::AlphaBlend_12bpp_12bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize, TUint8 aAlpha)
{
	TUint16*		dst_line = (TUint16*) TGameBitmapUtil::PixelAddress(this, aDstOrigin);
	const TUint16*	src_line = (TUint16*) TGameBitmapUtil::PixelAddress(&aSrcDevice, aSrcOrigin);

	TInt	height = aSize.iHeight;
	TInt	width = aSize.iWidth;

	for(TInt h=0 ; h<height ; h++)
	{
		TUint16*		dst = dst_line;
		const TUint16*	src = src_line;
		for (TInt w=0 ; w<width ; w++)
		{
			// alpha blending algorithm "borrowed" from MicroWindows ;-)
			unsigned int s = *src++;
			unsigned int d = *dst;
			unsigned int t = d & 0xf00;
			unsigned int m1, m2, m3;
			m1 = (((((s & 0xf00) - t)*aAlpha)>>8) & 0xf00) + t;
			t = d & 0x0f0;
			m2 = (((((s & 0xf0) - t)*aAlpha)>>8) & 0xf0) + t;
			t = d & 0xf;
			m3 = (((((s & 0xf) - t)*aAlpha)>>8) & 0xf) + t;
			*dst++ = (TUint16) (m1 | m2 | m3);
		}
		dst_line += (BytesPerLine() >> 1);
		src_line += (aSrcDevice.BytesPerLine() >> 1);
	}
}



void CGameScreenDevice::ScaledScanLineDraw_12or16bpp_8bpp(const TScaledBitmapContext& aScaledBitmapContext, TUint* aSrcLine, TBool aMerge)
{
	TInt		pixelCount = 0;
	TUint8*		src = (TUint8*) aSrcLine;
	TUint16*	dst = (TUint16*) aScaledBitmapContext.ScanLine();
	TUint*		pal = aScaledBitmapContext.SrcGameScreenDevice().Palette()->PaletteBuffer();
	TInt		width = aScaledBitmapContext.SrcGameScreenDevice().Size().iWidth;

	for (TInt w=0 ; w<width ; w++)
	{
		if (pixelCount == aScaledBitmapContext.SkipX())
		{
			pixelCount = 0;
			*(dst -1) |= pal[*src];
		}
		else
		{
			if (aMerge)
			{
				*dst++ |= pal[*src];
			}
			else
			{
				*dst++ = (TInt16) pal[*src];
			}
			pixelCount++;
		}
		src++;
	}
}



/*****************************************
 *
 * CFbsGameScreenDevice implementation
 *
 *****************************************/

CFbsGameScreenDevice::CFbsGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode) : CGameScreenDevice(aSize, aDisplayMode)
{
	iBytesPerLine = ((TGameBitmapUtil::GetBpp(aDisplayMode) * aSize.iWidth ) + 7) / 8;
}

CFbsGameScreenDevice::~CFbsGameScreenDevice()
{
	delete iBitmapContext;
	iBitmapContext = NULL;
	delete iBitmapDevice;
	iBitmapDevice = NULL;
	delete iBitmap;
	iBitmap = NULL;
}


CFbsGameScreenDevice* CFbsGameScreenDevice::NewL(const TSize& aSize, TDisplayMode aDisplayMode)
{
	CFbsGameScreenDevice*	self = new(ELeave) CFbsGameScreenDevice(aSize, aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

TUint8* CFbsGameScreenDevice::ScanLineAddress(TInt aLine) const
{
//	return (TUint8 *) iBitwiseBitmap->ScanLineAddress(iDataAddress, aLine);
	return (TUint8 *) iBitwiseBitmap->ScanLineAddress(iBitmap->DataAddress(), aLine);
}


void CFbsGameScreenDevice::ConstructL()
{
	iBitmap = new (ELeave) CExtraFbsBitmap;
	iBitmap->Create(iSize, iDisplayMode);

	iBitmapDevice = CFbsBitmapDevice::NewL(iBitmap);
	iBitmapDevice->CreateContext(iBitmapContext);

	iDataAddress = iBitmap->DataAddress();
	iBitwiseBitmap =iBitmap->BitmapAdressAddress();
}





/*****************************************
 *
 * CHwFbGameScreenDevice implementation
 *
 *****************************************/

CHwFbGameScreenDevice* CHwFbGameScreenDevice::NewL(const TSize& aWindowSize, TDisplayMode aDisplayMode)
{
	TScreenInfoV01			screenInfo;
	TPckg<TScreenInfoV01>	sI(screenInfo);
	UserSvr::ScreenInfo(sI);

	if (!screenInfo.iScreenAddressValid)
		User::Leave(KErrNotSupported);

		/*
		 * Note that aWindowSize may differ from screenInfo.iScreenSize.
		 * screenInfo.iScreenSize reflects the video controller's configured pixel dimensions, 
		 * whereas aWindowSize is the LCD's screen's maximim dimension.
		 */

	//DEB_Print(aWindowSize.iWidth, aWindowSize.iHeight, screenInfo.iScreenSize.iWidth, screenInfo.iScreenSize.iHeight);

	CHwFbGameScreenDevice*	self = new(ELeave) CHwFbGameScreenDevice(aWindowSize, aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL(screenInfo);
	CleanupStack::Pop();
	return self;
}


CHwFbGameScreenDevice::CHwFbGameScreenDevice(const TSize& aWindowSize, TDisplayMode aDisplayMode) : CGameScreenDevice(aWindowSize, aDisplayMode)
{
}

void CHwFbGameScreenDevice::ConstructL(const TScreenInfoV01& aScreenInfo)
{
	iFrameBuffer = (TUint8*) aScreenInfo.iScreenAddress;
	iBytesPerLine = ((TGameBitmapUtil::GetBpp(iDisplayMode) * aScreenInfo.iScreenSize.iWidth ) + 7) / 8;

	//DEB_Print((int)iFrameBuffer, iBytesPerLine, 0, 0);

//	RDebug::Print(_L("framebuffer=0x%x"), iFrameBuffer);

#ifdef __ER5__
		/*
		 * Hmm, I've just checked out the SA1100 spec. It seems that for 8bpp displays
		 * there is a 512 palette table at the beginning of the frame buffer.
		 * This code assumes that all ER5 target build with 8bpp are SA1100 machines. 
		 * I only know 2: Series 7 and Netbook.
		 */
	if (iDisplayMode == EColor8)
		iFrameBuffer += 512;
#endif
}


TUint8* CHwFbGameScreenDevice::ScanLineAddress(TInt aLine) const
{
	return iFrameBuffer + (BytesPerLine() * aLine);
}


/*****************************************
 *
 * CMemGameScreenDevice implementation
 *
 *****************************************/

CMemGameScreenDevice* CMemGameScreenDevice::NewL(const TSize& aSize, TDisplayMode aDisplayMode, TUint8* aMemBuffer)
{
	CMemGameScreenDevice*	self = new(ELeave) CMemGameScreenDevice(aSize, aDisplayMode, aMemBuffer);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}


CMemGameScreenDevice::CMemGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode, TUint8* aMemBuffer) : CGameScreenDevice(aSize, aDisplayMode), iMemBuffer(aMemBuffer)
{
	iOwnsMemBuf = (iMemBuffer == NULL);
	iBytesPerLine = ((TGameBitmapUtil::GetBpp(aDisplayMode) * aSize.iWidth ) + 7) / 8;
}


CMemGameScreenDevice::~CMemGameScreenDevice()
{
	if (iOwnsMemBuf)
	{
		User::Free(iMemBuffer);
		iMemBuffer = NULL;
	}
}

void CMemGameScreenDevice::ConstructL()
{
	if (iOwnsMemBuf)
	{
		TInt	size = iBytesPerLine*iSize.iHeight;
		iMemBuffer = (TUint8*) User::AllocL(size);
		Mem::FillZ(iMemBuffer, size);
	}
}


TUint8* CMemGameScreenDevice::ScanLineAddress(TInt aLine) const
{
	return iMemBuffer + (BytesPerLine() * aLine);
}


/***********************************************
 *
 * CVirtualKeyboard implementation
 *
 ***********************************************/


const TInt KVkRows = 5;
const TInt KVkCols = 12;

const TInt KKeyCodeSym		= 2;
const TInt KKeyCodeAlphaInc	= 4;
const TInt KKeyCodeAlphaDec	= 5;

const TInt KVkXMargin	= 1;
const TInt KVkYMargin	= 1;
const TInt KVkStandardWidth		= 16;
const TInt KVkStandardHeight	= 16;

#define IS_VALID_VIRTUAL_KEY(vk) ((vk)->iWidth)

const TVirtualKey	iDefaultVirtualKeyboardTable[KVkRows][KVkCols] = 
{
	{	/*                                         SHIFT  
		  width           scan    label  key    label  key
                          code          code          code */
		{KVkStandardWidth,   2,     "1", '1',     "!", '!'},
		{KVkStandardWidth,   3,     "2", '2',     "@", '@'},
		{KVkStandardWidth,   4,     "3", '3',     "#", '#'},
		{KVkStandardWidth,   5,     "4", '4',     "$", '$'},
		{KVkStandardWidth,   6,     "5", '5',     "%", '%'},
		{KVkStandardWidth,   7,     "6", '6',     "^", '^'},
		{KVkStandardWidth,   8,     "7", '7',     "&", '&'},
		{KVkStandardWidth,   9,     "8", '8',     "*", '*'},
		{KVkStandardWidth,  10,     "9", '9',     "(", '('},
		{KVkStandardWidth,  11,     "0", '0',     ")", ')'},
		{KVkStandardWidth + (KVkStandardWidth/2),   EStdKeyBackspace,     "<-", EKeyBackspace,     "<-", EKeyBackspace}
	},
	{
		{KVkStandardWidth + (KVkStandardWidth/2),  EStdKeyTab,     "tab", EKeyTab,     "TAB", EKeyTab},
		{KVkStandardWidth,  16,     "q", 'q',     "Q", 'Q'},
		{KVkStandardWidth,  17,     "w", 'w',     "W", 'W'},
		{KVkStandardWidth,  18,     "e", 'e',     "E", 'E'},
		{KVkStandardWidth,  19,     "r", 'r',     "R", 'R'},
		{KVkStandardWidth,  20,     "t", 't',     "T", 'T'},
		{KVkStandardWidth,  21,     "y", 'y',     "Y", 'Y'},
		{KVkStandardWidth,  22,     "u", 'u',     "U", 'U'},
		{KVkStandardWidth,  23,     "i", 'i',     "I", 'I'},
		{KVkStandardWidth,  24,     "o", 'o',     "O", 'O'},
		{KVkStandardWidth,  25,     "p", 'p',     "P", 'P'}
	},
	{
		{KVkStandardWidth + (KVkStandardWidth/2),  EStdKeyCapsLock,     "cps", EKeyCapsLock,     "CPS", EKeyCapsLock},
		{KVkStandardWidth,  30,     "a", 'a',     "A", 'A'},
		{KVkStandardWidth,  31,     "s", 's',     "S", 'S'},
		{KVkStandardWidth,  32,     "d", 'd',     "D", 'D'},
		{KVkStandardWidth,  33,     "f", 'f',     "F", 'F'},
		{KVkStandardWidth,  34,     "g", 'g',     "G", 'G'},
		{KVkStandardWidth,  35,     "h", 'h',     "H", 'H'},
		{KVkStandardWidth,  36,     "j", 'j',     "J", 'J'},
		{KVkStandardWidth,  37,     "k", 'k',     "K", 'K'},
		{KVkStandardWidth,  38,     "l", 'l',     "L", 'L'},
		{KVkStandardWidth,  59,     "@", KKeyCodeSym,     "@", KKeyCodeSym}		// 59 == scan code F1
	},
	{
		{KVkStandardWidth + (KVkStandardWidth/2),  EStdKeyLeftShift,     "sft", EKeyLeftShift,     "SFT", EKeyLeftShift},
		{KVkStandardWidth,  44,     "z", 'z',     "Z", 'Z'},
		{KVkStandardWidth,  45,     "x", 'x',     "X", 'X'},
		{KVkStandardWidth,  46,     "c", 'c',     "C", 'C'},
		{KVkStandardWidth,  47,     "v", 'v',     "V", 'V'},
		{KVkStandardWidth,  48,     "b", 'b',     "B", 'B'},
		{KVkStandardWidth,  49,     "n", 'n',     "N", 'N'},
		{KVkStandardWidth,  50,     "m", 'm',     "M", 'M'},
		{(KVkStandardWidth * 3) + (KVkXMargin*2), DEVICE_SELECT_SCANCODE,     "enter", DEVICE_SELECT_KEYCODE,     "ENTER", DEVICE_SELECT_KEYCODE},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0}
	},
	{
		{KVkStandardWidth + (KVkStandardWidth/2),  DEVICE_ESC_SCANCODE,     "esc", DEVICE_ESC_KEYCODE,     "ESC", DEVICE_ESC_KEYCODE},
		{(KVkStandardWidth * 3) + (KVkXMargin*2), 60,     "---", KKeyCodeAlphaDec,     "---", KKeyCodeAlphaDec},
		{(KVkStandardWidth * 3) + (KVkXMargin*2), EStdKeySpace,     "space", ' ',     "space", ' '},
		{(KVkStandardWidth * 3) + (KVkXMargin*2), 61,     "+++", KKeyCodeAlphaInc,     "+++", KKeyCodeAlphaInc},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0},
		{0, 0,     NULL, 0,     NULL, 0}
	}
};


CVirtualKeyboard::CVirtualKeyboard(CGameWindow& aGameWindow, CGameController& aGameController) : 
	iRect(TSize(200, 100)), iGameWindow(aGameWindow), iGameController(aGameController),
	iActiveRect(TSize(-1, -1))
{
	iBgColor =KRgbWhite;
	iFgColor =KRgbRed;
}

CVirtualKeyboard::~CVirtualKeyboard()
{
	if (iFont)
		iBitmap->FbsBitmapDevice()->ReleaseFont(iFont);
	if (iBitmap)
		delete iBitmap;
}


CVirtualKeyboard* CVirtualKeyboard::NewL(CGameWindow& aGameWindow, CGameController& aGameController)
{
	CVirtualKeyboard*	self = new(ELeave)CVirtualKeyboard(aGameWindow, aGameController);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}



void CVirtualKeyboard::ConstructL()
{
	iBitmap = CFbsGameScreenDevice::NewL(iRect.Size(), iGameWindow.GameScreenDevice()->DisplayMode());

	TFontSpec fontSpec(KFontName, 20);
	User::LeaveIfError(iBitmap->FbsBitmapDevice()->GetNearestFontInPixels(iFont, fontSpec));

	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->UseFont(iFont);

	// sothat we can detect our hotkey
	iGameController.AddGameControlInputHandlerL(*this);
	iGameWindow.AddScreenOrientationObserverL(*this);
	iGameController.AddGameControlRawInputHandlerL(*this);
}

void CVirtualKeyboard::SetPosition(const TPoint& aPosition)
{
	TSize	curSize = iRect.Size();
	TSize	screenSize = iGameWindow.GameScreenDevice()->Size();
	TPoint newPos = aPosition;
	if (newPos.iX < 0)
		newPos.iX = 0;
	if (newPos.iX + curSize.iWidth > screenSize.iWidth)
		newPos.iX = screenSize.iWidth - curSize.iWidth;
	if (newPos.iY < 0)
		newPos.iY = 0;
	if (newPos.iY + curSize.iHeight> screenSize.iHeight)
		newPos.iY = screenSize.iHeight - curSize.iHeight;

	iRect = TRect(newPos, curSize);
}


void CVirtualKeyboard::ScreenOrientationChanged(TOrientation aOrientation)
{
	iOrientation = aOrientation;

	DrawKeyboard();
}


void CVirtualKeyboard::CheckForVirtualKeyboardHotKey(TInt aKeyCode)
{
	if (IsHotKeyForVirtualKeyboard(aKeyCode))
	{
		if (!iIsActive)
			ActivateL();
		else
			Deactivate();
	}
}


void CVirtualKeyboard::ActivateL()
{
	iIsActive = ETrue;
	iAlphaInc = 0;
	iAlpha = 175;
	iKey = NULL;

	iGameWindow.AddGameScreenDrawObserverL(*this);

	DrawKeyboard();
}


void CVirtualKeyboard::Deactivate()
{
	iIsActive = EFalse;

	// no longer need to steal pointer events
	iGameWindow.RemoveGameScreenDrawObserver(*this);
	iKey = NULL;
	ClearKeyboard();
}


void CVirtualKeyboard::PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice)
{
	if (iAlphaInc)
	{
		iAlpha += iAlphaInc;
		if (iAlpha > 255)
			iAlpha = 255;
		if (iAlpha < 0)
			iAlpha = 0;
	}

	aDstGameScreenDevice->AlphaBlend(iRect.iTl, *iBitmap, TRect(iBitmap->Size()), (TUint8) iAlpha);
}

TBool CVirtualKeyboard::WantDrawToScreen()
{
	return iIsActive;
}

void CVirtualKeyboard::HandleVirtualButtonDown(const TVirtualButton* aVirtualButton)
{
	iGameController.InjectkeyEvent(EEventKeyDown, ScanCode(aVirtualButton));
	iGameController.InjectkeyEvent(EEventKey, KeyCode(aVirtualButton));
	iButton = aVirtualButton;
}

void CVirtualKeyboard::HandleVirtualButtonUp()
{
	if (iButton)
	{
		iGameController.InjectkeyEvent(EEventKeyUp, ScanCode(iButton));
		iButton = NULL;
	}
}


TInt CVirtualKeyboard::KeyCode(const TVirtualButton* aVirtualButton)
{
	return aVirtualButton->iKeyCode;
}

TInt CVirtualKeyboard::ScanCode(const TVirtualButton* aVirtualButton)
{
	return iKeyCodesAreScanCodes ? aVirtualButton->iKeyCode : aVirtualButton->iScanCode;
}

const TVirtualButton* CVirtualKeyboard::FindVirtualButton(const TPoint& aPoint)
{
	const TVirtualButton*	button = &iVirtualButtons[0];
	while((button->iScanCode != -1) && (button->iKeyCode != -1))
	{
		TRect	rect(button->iTlX, button->iTlY, button->iBrX, button->iBrY);
		if (rect.Contains(aPoint))
			return button;
		button++;
	}
	return NULL;
}


TBool CVirtualKeyboard::CheckVirtualButton(const TPointerEvent& aPointerEvent)
{
	TBool	consumed = EFalse;

	if (aPointerEvent.iType == TPointerEvent::EButton1Down)
	{
		const TVirtualButton* virtualButton = FindVirtualButton(aPointerEvent.iPosition);
		if (virtualButton)
		{
			HandleVirtualButtonDown(virtualButton);
			consumed = ETrue;
		}
	}
	else if (aPointerEvent.iType == TPointerEvent::EButton1Up)
	{
		if (iButton)
		{
			HandleVirtualButtonUp();
			consumed = ETrue;
		}
	}
	return consumed;
}


TBool CVirtualKeyboard::ConsumeRawPointerEvent(const TPointerEvent& aPointerEvent)
{
//	RDebug::Print(_L("CVirtualKeyboard::ConsumeRawPointerEvent, type=%d, (%d, %d)"), aPointerEvent.iType, aPointerEvent.iPosition.iX, aPointerEvent.iPosition.iY);

	TPoint	p = aPointerEvent.iPosition;
	TBool	consumed = EFalse;
	if (iIsActive && iRect.Contains(p))
	{
		consumed = ETrue;
		if (aPointerEvent.iType == TPointerEvent::EButton1Down)
		{
			const TVirtualKey* virtualKey = FindVirtualKey(p - iRect.iTl);
			if (virtualKey)
				HandleVirtualKeyDown(virtualKey);
		}
		else if (aPointerEvent.iType == TPointerEvent::EButton1Up)
		{
			HandleVirtualKeyUp();
		}
	}
	else
	{
		if (iVirtualButtons)
			consumed = CheckVirtualButton(aPointerEvent);
	}
//	RDebug::Print(_L("CVirtualKeyboard::ConsumeRawPointerEvent, consume=%d)"), consumed);

	return consumed;
}


TInt CVirtualKeyboard::KeyCode(const TVirtualKey* aVirtualKey)
{
	TBool shift = iCapsMode ^ iShiftMode;
	return shift? aVirtualKey->iShiftKeyCode: aVirtualKey->iKeyCode;
}

TInt CVirtualKeyboard::ScanCode(const TVirtualKey* aVirtualKey)
{
	return iKeyCodesAreScanCodes ? aVirtualKey->iKeyCode : aVirtualKey->iScanCode;
}

void CVirtualKeyboard::HandleVirtualKeyDown(const TVirtualKey* aVirtualKey)
{
	TBool generateKeyEvent = !((aVirtualKey->iKeyCode == EKeyLeftShift) 
								|| (aVirtualKey->iKeyCode == EKeyCapsLock) || 
								(aVirtualKey->iKeyCode == KKeyCodeSym));
	if (generateKeyEvent)
	{
		iGameController.InjectkeyEvent(EEventKeyDown, ScanCode(aVirtualKey));
		iGameController.InjectkeyEvent(EEventKey, KeyCode(aVirtualKey));
	}
	iKey = aVirtualKey;
	DrawKeyboard();

	if (aVirtualKey->iKeyCode == KKeyCodeAlphaInc)
		iAlphaInc = 1;
	else if (aVirtualKey->iKeyCode == KKeyCodeAlphaDec)
		iAlphaInc = -1;
}

void CVirtualKeyboard::HandleVirtualKeyUp()
{
	if (iKey)
	{
		if (iKey->iKeyCode == EKeyLeftShift)
			iShiftMode = !iShiftMode;
		else if (iKey->iKeyCode == EKeyCapsLock)
			iCapsMode = !iCapsMode;
		else
		{
			iGameController.InjectkeyEvent(EEventKeyUp, ScanCode(iKey));
			iKey = NULL;
			if (iShiftMode)
				iShiftMode = EFalse;
		}
		DrawKeyboard();
	}
	iAlphaInc = 0;
}



void CVirtualKeyboard::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
	if (aType == EEventKey)
		CheckForVirtualKeyboardHotKey(aKeyEvent.iCode);
}

void CVirtualKeyboard::HandlePointerEvent(const TPointerEvent& aPointerEvent)
{
	if ((aPointerEvent.iType == TPointerEvent::EButton1Down) && iActiveRect.Contains(aPointerEvent.iPosition))
	{
		if (!iIsActive)
			ActivateL();
		else
			Deactivate();
	}
}

const TVirtualKey* CVirtualKeyboard::FindVirtualKey(const TPoint& aPoint)
{
	TInt	row = aPoint.iY / (KVkStandardHeight + KVkYMargin);
	TPoint	p(KVkXMargin, row * (KVkStandardHeight + KVkYMargin));
	if (row>=0 && row<KVkRows)
	{
		for (TInt col = 0; col<KVkCols ; col++)
		{
			const TVirtualKey& vk = iDefaultVirtualKeyboardTable[row][col];
			if (IS_VALID_VIRTUAL_KEY(&vk))
			{
				TRect	rect(p, p + TPoint (vk.iWidth, KVkStandardHeight));
				if (rect.Contains(aPoint))
					return &vk;
			}
			p.iX += KVkXMargin + vk.iWidth;
		}
	}
	return NULL;
}

void CVirtualKeyboard::ClearKeyboard()
{
	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->SetBrushColor(KRgbBlack);
	gc->SetBrushStyle(CGraphicsContext::ENullBrush);
	gc->Clear();
}


CFbsBitGc::TGraphicsOrientation ToGraphicsOrientation(TOrientation aOrientation)
{
	static const CFbsBitGc::TGraphicsOrientation	orientatationList[] = 
	{
		CFbsBitGc::EGraphicsOrientationNormal,
		CFbsBitGc::EGraphicsOrientationRotated90,
		CFbsBitGc::EGraphicsOrientationRotated270,
		CFbsBitGc::EGraphicsOrientationRotated180
	};
	return orientatationList[aOrientation];
}

void CVirtualKeyboard::DrawKeyboard()
{
	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->SetBrushColor(iBgColor);
	gc->SetBrushStyle(CGraphicsContext::ENullBrush);
	gc->Clear();
	gc->SetPenColor(iFgColor);
	gc->SetPenSize(TSize(1,1));
	gc->UseFont(iFont);

	TBool shift = iCapsMode ^ iShiftMode;
	TPoint	p(KVkYMargin, KVkXMargin);
	for (TInt row = 0; row<KVkRows ; row++)
	{
		p.iX = KVkXMargin;
		for (TInt col = 0; col<KVkCols ; col++)
		{
			const TVirtualKey& vk = iDefaultVirtualKeyboardTable[row][col];
			if (IS_VALID_VIRTUAL_KEY(&vk))
			{
				TRect	rect(p, p + TPoint (vk.iWidth, KVkStandardHeight));
				if (vk.iText)
				{
					TBool isReversed = ((vk.iKeyCode == EKeyLeftShift) && (iShiftMode))
							|| ((vk.iKeyCode == EKeyCapsLock) && (iCapsMode));
					if (iKey == &vk)
						isReversed = ETrue;

					if (isReversed)
					{
						gc->SetBrushColor(iFgColor);
						gc->SetPenColor(iBgColor);
						gc->Clear(rect);
					}
					
					// draw border
					gc->DrawRect(rect);

					// draw text
					TPtrC8		text8((TUint8*) (shift ? vk.iShiftText : vk.iText));
					TBuf<32>	text;
					text.Copy(text8);
					TInt baselineOffset=(rect.Height()+iFont->AscentInPixels())/2; 
					gc->DrawText(text, rect, baselineOffset, CGraphicsContext::ECenter);

					if (isReversed)
					{
						gc->SetBrushColor(iBgColor);
						gc->SetPenColor(iFgColor);
					}
				}
			}
			p.iX += KVkXMargin + vk.iWidth;
		}
		p.iY += KVkYMargin + KVkStandardHeight;
	}
	gc->SetOrientation(ToGraphicsOrientation(iOrientation));
}




/***********************************************
 *
 * CGameWindow
 *
 ***********************************************/


CGameWindow::CGameWindow()
{
}


CGameWindow::~CGameWindow()
{
	//delete iGameIntro;	// MR
	iGameIntro = NULL;

	delete iInfoMgr;
	iInfoMgr = NULL;

	delete iSmartPenToKeyInputHandler;
	iSmartPenToKeyInputHandler = 0;

	//DestructVirtualKeyboard();	// MR
	DestructGameController();
	DestructScreenDevice();
	DestructWsResources();

	delete iScreenOrientationObservers;
	iScreenOrientationObservers = NULL;

	delete iGameScreenDrawObservers;
	iGameScreenDrawObservers = NULL;
}


CGameWindow* CGameWindow::NewL(const TDesC& aGameName)
{
	CGameWindow*	self = new(ELeave)CGameWindow;
	CleanupStack::PushL(self);
	self->ConstructL(aGameName);
	CleanupStack::Pop();
	return self;
}


void CGameWindow::ConstructL(const TDesC& aGameName)
{
	iGameName = aGameName.Left(iGameName.MaxLength());

	iGameScreenDrawObservers = new(ELeave)CArrayPtrFlat<MGameScreenDrawObserver>(5);
	iScreenOrientationObservers = new(ELeave)CArrayPtrFlat<MScreenOrientationObserver>(5);

	ConstructWsResourcesL();
	ConstructScreenDeviceL();
	ConstructGameControllerL();
	//ConstructVirtualKeyboardL();	// MR
	ClearScreen();

	SetGameScreenDrawer(&iDirectBitmapDrawer);
	iGamePaner.ActivateL(this, iGameController, EOrientationNormal);

	iInfoMgr = CInfoMgr::NewL(*this, *iGameController);
	//iGameIntro = CGameIntro::NewL(*this, *iGameController);	// MR

	iMemoryGameScreenDevice = 	CMemGameScreenDevice::NewL(iGameScreenDevice->Size(), iGameScreenDevice->DisplayMode());
	CaptureKeyEvents();
}

void CGameWindow::ConstructSmartPenToKeyInputHandlerL()
{
	if (!iSmartPenToKeyInputHandler)
	{
		iSmartPenToKeyInputHandler = CSmartPenToKeyInputHandler::NewL(*this, *iGameController);
	}
}

void CGameWindow::CanDo_X_And_Y_AtTheSameTime(TBool aCanDo)
{
	if (iSmartPenToKeyInputHandler)
		iSmartPenToKeyInputHandler->CanDo_X_And_Y_AtTheSameTime(aCanDo);
}



void CGameWindow::CaptureKeyEvents()
{
	// MR : /*, 1*/
	if (!iCapturedButtons[0])
		iCapturedButtons[0] = iWsWindowGroup.CaptureKey(EKeyApplication0, 0, 0/*, 1*/);
	if (!iCapturedButtons[1])
		iCapturedButtons[1] = iWsWindowGroup.CaptureKey(EKeyApplication1, 0, 0/*, 1*/);
	if (!iCapturedButtons[2])
		iCapturedButtons[2] = iWsWindowGroup.CaptureKey(EQuartzKeyBack, 0, 0/*, 1*/);
	if (!iCapturedButtons[3])
		iCapturedButtons[3] = iWsWindowGroup.CaptureKey(EQuartzKeyMenu2, 0, 0/*, 1*/);
	if (!iCapturedButtons[4])
		iCapturedButtons[4] = iWsWindowGroup.CaptureKey(EKeyDevice0, 0, 0/*, 1*/);
}

void CGameWindow::ReleaseKeyEvents()
{
	for (TInt i=0 ; i<KMaxCapturedButtons ; i++)
	{
		if (iCapturedButtons[i])
		{
			iWsWindowGroup.CancelCaptureKey(iCapturedButtons[i]);
			iCapturedButtons[i] = 0;
		}
	}
}

void CGameWindow::Hide()
{
	ReleaseKeyEvents();
	iWsWindowGroup.SetOrdinalPosition(KMinTInt, KMinTInt);
	iWsWindow.SetVisible(EFalse);

	TRect	rect(iWsWindow.Position(), iWsWindow.Size());
	iWindowGc->Activate(iWsWindow);
	iWsWindow.Invalidate();
	iWsWindow.BeginRedraw(rect);
	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	
	iWsSession.Flush();
	iWasPaused = ETrue;
}

void CGameWindow::Show()
{
	iWsWindowGroup.SetOrdinalPosition(0, 0);
	iWsWindow.SetVisible(ETrue);
	CaptureKeyEvents();
}


void CGameWindow::ConstructWsResourcesL()
{
	TInt	error;

	error = iWsSession.Connect();
	User::LeaveIfError(error);
	iWsScreen=new(ELeave) CWsScreenDevice(iWsSession);
	User::LeaveIfError(iWsScreen->Construct());
	User::LeaveIfError(iWsScreen->CreateContext(iWindowGc));

	iWsWindowGroup=RWindowGroup(iWsSession);
	User::LeaveIfError(iWsWindowGroup.Construct((TUint32)&iWsWindowGroup));
	iWsWindowGroup.SetOrdinalPosition(0);
	iWsWindowGroup.SetName(iGameName);

	iWsWindow=RWindow(iWsSession);
	User::LeaveIfError(iWsWindow.Construct(iWsWindowGroup, (TUint32)&iWsWindow));
	iWsWindow.Activate();
	iWsWindow.SetSize(iWsScreen->SizeInPixels());

#ifdef __TEST_DIRECT_SCREEN_ACCESS__
	// you can then click on the app picker to force something to move in front of the game 
	iWsWindow.SetPosition(TPoint(0, 4));
#endif

	iWsWindow.SetVisible(ETrue);
	iWsWindow.SetPointerGrab(ETrue);
	iWsWindow.PointerFilter(EPointerFilterDrag, 0);

		/*
		 * Tell the Window Server not to mess about with our process priority
		 * Also, because of the way legacy games are written, they never sleep
		 * and thus never voluntarily yield the CPU. We set our process priority
		 * to EPriorityForeground and hope that a Telephony application on
		 * this device runs at EPriorityForeground as well. If not, tough! ;-)
		 */
	iWsSession.ComputeMode(RWsSession::EPriorityControlDisabled);
	RProcess		me;
	me.SetPriority(EPriorityForeground);
}


void CGameWindow::DestructWsResources()
{
	if (iWsWindow.WsHandle())
		iWsWindow.Close();

	if (iWsWindowGroup.WsHandle())
		iWsWindowGroup.Close();

	delete iWindowGc;
	iWindowGc = NULL;

	delete iWsScreen;
	iWsScreen = NULL;

	if (iWsSession.WsHandle())
		iWsSession.Close();
}


void CGameWindow::ConstructGameControllerL()
{
	iGameController = CGameController::NewL(iWsSession);
	AddGameScreenDrawObserverL(*iGameController);
	AddScreenOrientationObserverL(*iGameController);
}

void CGameWindow::DestructGameController()
{
	delete iGameController;
	iGameController = NULL;
}

void CGameWindow::ConstructScreenDeviceL()
{
	TDisplayMode	mode = iWsScreen->DisplayMode();
	TSize			size = iWsScreen->SizeInPixels();

	//DEB_Print(size.iWidth, size.iHeight, mode, 0);

//#ifdef __WINS__
#if 1	// MR: CHwFbGameScreenDevice a chie sur le SX1
	iUseWServ = ETrue;
#endif

	if (iUseWServ)
	{
			/*
			 * For WINS we emulate using a Fbs bitmap which we blit to the window's device
			 * context. You can test most behaviour by:
			 * 1) adjusting the size and display mode in the CFbsGameScreenDevice::NewL
			 * 2) configuring the emulator for the appropriate size and mode
			 */
		iGameScreenDevice = CFbsGameScreenDevice::NewL(size, mode);
	}
	else
	{
		iGameScreenDevice = CHwFbGameScreenDevice::NewL(size, mode);
	}

	iGameDirectScreenAccess = CGameDirectScreenAccess::NewL(*this);
	AddGameScreenDrawObserverL(*iGameDirectScreenAccess);
}


void CGameWindow::DestructScreenDevice()
{
	delete iGameScreenDevice;
	iGameScreenDevice = NULL;

	delete iGameDirectScreenAccess;
	iGameDirectScreenAccess = NULL;
}


void CGameWindow::ConstructVirtualKeyboardL()
{
	iVirtualKeyboard = CVirtualKeyboard::NewL(*this, *iGameController);
}

void CGameWindow::DestructVirtualKeyboard()
{
	delete iVirtualKeyboard;
	iVirtualKeyboard = NULL;
}

void CGameWindow::GameFrameTick(CGameScreenDevice* aSrcGameScreenDevice)
{
	TBool				useMemoryScreenDevice = EFalse;
	CGameScreenDevice*	screenDevice = iGameScreenDevice;

		/*
		 * If any of the screen draw observers want to draw something to screen
		 * we will use the memory screen device first. This will prevent the usual
		 * screen flicker.
		 * Also, if we used to use the memory screen device and this frame we don't
		 * then we use it one more time to properly clear the screen.
		 */
	for (TInt ii=0 ; ii<iGameScreenDrawObservers->Count() && !useMemoryScreenDevice; ii++)
		useMemoryScreenDevice = (*iGameScreenDrawObservers)[ii]->WantDrawToScreen();

	if (useMemoryScreenDevice || iUsedMemoryScreenDevice || iWasPaused)
	{
		screenDevice = iMemoryGameScreenDevice;
		iMemoryGameScreenDevice->Clear();
		iWasPaused = EFalse;
	}

	for (TInt i=0 ; i<iGameScreenDrawObservers->Count() ; i++)
		(*iGameScreenDrawObservers)[i]->PreDrawFrameToScreen(screenDevice);

	if (iGameScreenDrawer)
		iGameScreenDrawer->DrawFrame(screenDevice, aSrcGameScreenDevice);

	for (TInt j=0 ; j<iGameScreenDrawObservers->Count() ; j++)
		(*iGameScreenDrawObservers)[j]->PostDrawFrameToScreen(screenDevice);

	if (useMemoryScreenDevice || iUsedMemoryScreenDevice)
		iGameScreenDevice->DirectDraw(TPoint(0,0), *iMemoryGameScreenDevice, iGameScreenDevice->Size());

	iUsedMemoryScreenDevice = useMemoryScreenDevice;

	if (iUseWServ)
		WsBlitBitmap((CFbsGameScreenDevice*) iGameScreenDevice);

}

void CGameWindow::ClearScreen()
{
	iGameScreenDevice->Clear();

	if (iUseWServ)
		WsBlitBitmap((CFbsGameScreenDevice*) iGameScreenDevice, TRect(iGameScreenDevice->Size()));
}


void CGameWindow::SetGameScreenDrawer(MGameScreenDrawer* aGameScreenDrawer)
{
	if (aGameScreenDrawer)
		iGameScreenDrawer = aGameScreenDrawer;
	else
		iGameScreenDrawer = &iDirectBitmapDrawer;
}



void CGameWindow::WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice)
{
	WsBlitBitmap(aFbsGameScreenDevice, GameVisibleArea());
}

void CGameWindow::WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice, const TRect& aRect)
{
	iWindowGc->Activate(iWsWindow);

	iWsWindow.Invalidate(aRect);
	iWsWindow.BeginRedraw(aRect);

	if (aFbsGameScreenDevice)
	{
		iWindowGc->BitBlt(aRect.iTl, aFbsGameScreenDevice->FbsBitmap(), aRect);
	}

	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	iWsSession.Flush();
}


void CGameWindow::AddScreenOrientationObserverL(MScreenOrientationObserver& aObserver)
{
	iScreenOrientationObservers->AppendL(&aObserver);
	aObserver.ScreenOrientationChanged(iOrientation);
}

void CGameWindow::SetScreenOrientation(TOrientation aOrientation)
{
	iOrientation = aOrientation;
	for (TInt i=0 ; i<iScreenOrientationObservers->Count() ; i++)
		(*iScreenOrientationObservers)[i]->ScreenOrientationChanged(iOrientation);
}



/*****************************************
 *
 * CGameController
 *
 *****************************************/

CGameController* CGameController::NewL(RWsSession& aWsSession)
{
	CGameController*	self = new(ELeave)CGameController(aWsSession);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

CGameController::~CGameController()
{
	delete iGameControlInputHandlers;
	delete iGameControlRawInputHandlers;
}

static TGameScanCodeTable	sP800GameScanCodeTable[] = 
{
#ifndef PROJECT_EDOOM
	{DEVICE_LEFT_SCANCODE, GAME_BUTTON2_SCANCODE},
	{DEVICE_RIGHT_SCANCODE, GAME_BUTTON2_SCANCODE},
	{DEVICE_UP_SCANCODE, GAME_BUTTON1_SCANCODE},
	{DEVICE_DOWN_SCANCODE, GAME_BUTTON1_SCANCODE},
	{DEVICE_SELECT_SCANCODE, GAME_BUTTON2_SCANCODE},
#endif
	{0, 0}
};

static TGameKeyCodeTable	sP800GameKeyCodeTable[] = 
{
#ifndef PROJECT_EDOOM
	{DEVICE_LEFT_KEYCODE, GAME_BUTTON2_KEYCODE},
	{DEVICE_RIGHT_KEYCODE, GAME_BUTTON2_KEYCODE},
	{DEVICE_UP_KEYCODE, GAME_BUTTON1_KEYCODE},
	{DEVICE_DOWN_KEYCODE, GAME_BUTTON1_KEYCODE},
	{DEVICE_SELECT_KEYCODE, GAME_BUTTON2_KEYCODE},
#endif
	{0, 0}
};

void CGameController::ConstructL()
{
	iGameControlInputHandlers = new(ELeave)CArrayPtrFlat<MGameControlInputHandler>(5);
	iGameControlRawInputHandlers = new(ELeave)CArrayPtrFlat<MGameControlRawInputHandler>(5);

	SetGameKeymapping(sP800GameScanCodeTable[0], sP800GameKeyCodeTable[0]);

	StartWServEvents();
}

void CGameController::SetGameKeymapping(TGameScanCodeTable& aGameScanCodeTable, TGameKeyCodeTable& aGameKeyCodeTable)
{
	iGameScanCodeTable = &aGameScanCodeTable;
	iGameKeyCodeTable = &aGameKeyCodeTable;
}


void CGameController::StartWServEvents()
{
	iWsEventStatus = KRequestPending;
	iWsSession.EventReady(&iWsEventStatus);

	iRedrawEventStatus = KRequestPending;
	iWsSession.RedrawReady(&iRedrawEventStatus);
}

void CGameController::CancelWServEvents()
{
	if (iWsEventStatus != KRequestPending)
		iWsSession.EventReadyCancel();

	if (iRedrawEventStatus != KRequestPending)
		iWsSession.RedrawReadyCancel();
}


void CGameController::PollWServEvents()
{
	while (iWsEventStatus != KRequestPending)
	{
		iWsSession.GetEvent(iWsEvent);
		HandleWsEvent(iWsEvent);
		iWsEventStatus = KRequestPending;
		iWsSession.EventReady(&iWsEventStatus);
	}
}

void CGameController::HandleWsEvent(const TWsEvent& aWsEvent)
{
	if ((aWsEvent.Type() == EEventKeyDown) || (aWsEvent.Type() == EEventKey) || (aWsEvent.Type() == EEventKeyUp))
	{
		HandleKeyEvent(aWsEvent.Type(), *aWsEvent.Key());
	} 
	else if (aWsEvent.Type() == EEventPointer)
	{
		HandlePointerEvent(*aWsEvent.Pointer());
	}
	else
	{
//		RDebug::Print(_L("HandleWsEvent, event=%d"), aWsEvent.Type());
	}

}


void CGameController::ScreenOrientationChanged(TOrientation aOrientation)
{
	iOrientation = aOrientation;
}


void CGameController::PostDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/)
{
	PollWServEvents();
}

		/*
		 * Raw pointer events
		 */
void CGameController::HandlePointerEvent(const TPointerEvent& aPointerEvent)
{
//	RDebug::Print(_L("CGameController::HandlePointerEvent, type=%d, (%d, %d)"), aPointerEvent.iType, aPointerEvent.iPosition.iX, aPointerEvent.iPosition.iY);
	TBool	consumed = EFalse;
	TInt	count = iGameControlRawInputHandlers->Count();
	for (TInt i=count-1 ; i>=0 && !consumed ; i--)
		consumed=(*iGameControlRawInputHandlers)[i]->ConsumeRawPointerEvent(aPointerEvent);

	if (!consumed)
	{
		count = iGameControlInputHandlers->Count();
		for (i=0 ; i<count ; i++)
			(*iGameControlInputHandlers)[i]->HandlePointerEvent(aPointerEvent);
	}
}



void CGameController::MapToGameEvents(TInt aType, TKeyEvent& aKeyEvent)
{
	if ((aType == EEventKeyDown) || (aType == EEventKeyUp))
	{
//		RDebug::Print(_L("SCAN CODE=%d"), aKeyEvent.iScanCode);
		TGameScanCodeTable*		gameScanCodeEntry = iGameScanCodeTable;
		while (gameScanCodeEntry->iGameScanCode)
		{
			if (gameScanCodeEntry->iPhysicalScanCode == aKeyEvent.iScanCode)
			{
				aKeyEvent.iScanCode = gameScanCodeEntry->iGameScanCode;
				return;
			}
			gameScanCodeEntry++;
		}
	}
	else if (aType == EEventKey)
	{
//		RDebug::Print(_L("KEY CODE=%d"), aKeyEvent.iCode);
		TGameKeyCodeTable*		gameKeyCodeEntry = iGameKeyCodeTable;
		while (gameKeyCodeEntry->iGameKey)
		{
			if (gameKeyCodeEntry->iPhysicalKey == aKeyEvent.iCode)
			{
				aKeyEvent.iCode = gameKeyCodeEntry->iGameKey;
				return;
			}
			gameKeyCodeEntry++;
		}
	}
}


void CGameController::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
	TKeyEvent	 keyEvent = aKeyEvent;

	MapToGameEvents(aType, keyEvent);

	TBool	consumed = EFalse;
	TInt	count = iGameControlRawInputHandlers->Count();

	for (TInt i=count-1 ; i>=0 && !consumed ; i--)
	{	consumed=(*iGameControlRawInputHandlers)[i]->ConsumeRawKeyEvent(aType, keyEvent);
		if (consumed)
			(*iGameControlInputHandlers)[i]->HandleKeyEvent(aType, keyEvent);
	}

//	RDebug::Print(_L("NOT consumed start"));
	if (!consumed)
	{
		count = iGameControlInputHandlers->Count();
		for (i=0 ; i<count ; i++)
			(*iGameControlInputHandlers)[i]->HandleKeyEvent(aType, keyEvent);
	}
//	RDebug::Print(_L("NOT consumed end"));
}


void CGameController::InjectkeyEvent(TInt aEventType, TInt aKeyOrScanKey, TInt aRepeat)
{
	TKeyEvent	keyEvent;

	keyEvent.iCode = (aEventType == EEventKey) ? aKeyOrScanKey: 0;
	keyEvent.iScanCode = (aEventType != EEventKey) ? aKeyOrScanKey: 0;
	keyEvent.iModifiers = 0;
	keyEvent.iRepeats = aRepeat;

//	RDebug::Print(HERE("type=%d, key=%d"), aEventType, aKeyOrScanKey);

	HandleKeyEvent(aEventType, keyEvent);
}


/******************************************
 *
 * CMameStyleInputHandler implementation
 *
 ******************************************/

CMameStyleInputHandler::CMameStyleInputHandler()
{
}

TBool CMameStyleInputHandler::IsAnyKeyPressed()
{
	return (iNoOfKeyPresses > 0);
}

TBool CMameStyleInputHandler::IsKeyPressed(TInt aKeyCode)
{
#ifdef __WINS__
	RDebug::Print(_L("KEY PRESSED? key=%d"), aKeyCode);
#endif
	TInt presses = 0;

	for (TInt i=0 ; i<MAX_SIMULTANEOUS_KEYPRESSES && presses<iNoOfKeyPresses; i++)
	{
		if (iScanCodes[i])
		{
			presses++;
			if (iKeyCodes[i] == aKeyCode)
			{
				if (DelayKeyUpEvent(iScanCodes[i]))
				{
//					RDebug::Print(_L("Processed DelayKeyUpEvent, scan=%d"), iScanCodes[i]);
					iScanCodes[i] = 0;
					iKeyCodes[i] = 0;
					iNoOfKeyPresses--;
				}
				return ETrue;
			}
		}
	}
	if (iNoIsKeyPressed++ > 30)
	{
		iNoIsKeyPressed = 0;
		ForceDelayedKeyUpEvents();
	}

	return EFalse;
}

void CMameStyleInputHandler::ForceDelayedKeyUpEvents()
{
	TInt presses = 0;
	for (TInt i=0 ; i<MAX_SIMULTANEOUS_KEYPRESSES && presses<iNoOfKeyPresses; i++)
	{
		if (iScanCodes[i])
		{
			presses++;
			if (DelayKeyUpEvent(iScanCodes[i]))
			{
//				RDebug::Print(_L("Processed DelayKeyUpEvent, scan=%d"), iScanCodes[i]);
				iScanCodes[i] = 0;
				iKeyCodes[i] = 0;
				iNoOfKeyPresses--;
			}
		}
	}
}


void CMameStyleInputHandler::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
	if (aType ==EEventKeyDown)
	{
		iLastScanCode = aKeyEvent.iScanCode;
	}
	else if (aType == EEventKey)
	{
		if (aKeyEvent.iRepeats == 0)
			KeyDown(iLastScanCode, aKeyEvent.iCode);
	}
	else if (aType == EEventKeyUp)
	{
		KeyUp(aKeyEvent.iScanCode);
	}
}

void CMameStyleInputHandler::KeyDown(TInt aScanCode, TInt aKeyCode)
{
	if (iNoOfKeyPresses < MAX_SIMULTANEOUS_KEYPRESSES)
	{
		for (TInt i=0 ; i<MAX_SIMULTANEOUS_KEYPRESSES; i++)
		{
			if (iScanCodes[i] == 0)
			{
				iScanCodes[i] = aScanCode;
				iKeyCodes[i] = aKeyCode;
				iNoOfKeyPresses++;
				break;
			}
		}
	}
	else
	{
		ASSERT(EFalse);
	}
}


TBool CMameStyleInputHandler::DelayKeyUpEvent(TInt aScanCode)
{
	return (aScanCode == GAME_BUTTON1_SCANCODE);
}

void CMameStyleInputHandler::KeyUp(TInt aScanCode)
{
	if (DelayKeyUpEvent(aScanCode))
	{
//		RDebug::Print(_L("DelayKeyUpEvent, scan=%d"), aScanCode);
		return;
	}
	for (TInt i=0 ; i<MAX_SIMULTANEOUS_KEYPRESSES; i++)
	{
		if (iScanCodes[i] == aScanCode)
		{
			iScanCodes[i] = 0;
			iKeyCodes[i] = 0;
			iNoOfKeyPresses--;
			break;
		}
	}
}


/**********************************************
 *
 * TDirectBitmapDrawer
 *
 **********************************************/

void TDirectBitmapDrawer::DrawFrame(CGameScreenDevice* aDstGameScreenDevice, const CGameScreenDevice* aSrcGameScreenDevice)
{
	if (!iSetup)
	{
		iGameSize = aSrcGameScreenDevice->Size();

		TSize	screenSize = aDstGameScreenDevice->Size();
		TInt	diffX = screenSize.iWidth - iGameSize.Size().iWidth;
		TInt	diffY = screenSize.iHeight - iGameSize.Size().iHeight;

		iGameOrigin.iX = (diffX>0) ? diffX/2 : 0;
		iGameOrigin.iY = (diffY>0) ? diffY/2 : 0;

		iSetup = ETrue;
		iGameVisibleArea = TGameBitmapUtil::VisibleIntersection(screenSize, iGameOrigin, iGameSize.Size(), iGameSize);
	}

	aDstGameScreenDevice->DirectDraw(iGameOrigin, *aSrcGameScreenDevice, iGameSize);
}


TRect TDirectBitmapDrawer::GameVisibleArea() const
{
	return iGameVisibleArea;
}

TSize TDirectBitmapDrawer::GameSize() const
{
	return iGameSize.Size();
}

TPoint TDirectBitmapDrawer::GameOrigin() const
{
	return iGameOrigin;
}

void TDirectBitmapDrawer::SetGameOrigin(const TPoint& aOrigin)
{
	iGameOrigin = aOrigin;
}


/********************************************
 *
 * TScaledBitmapDrawer implementation
 *
 ********************************************/

void TScaledBitmapDrawer::DrawFrame(CGameScreenDevice* aDstGameScreenDevice, const CGameScreenDevice* aSrcGameScreenDevice)
{
	if (!iScaledBitmapContext.HasSetup())
		iScaledBitmapContext = TScaledBitmapContext(*aDstGameScreenDevice, *aSrcGameScreenDevice, aSrcGameScreenDevice->Size(), ETrue);
	aDstGameScreenDevice->ScaledDrawBitmap(iScaledBitmapContext);
}

TRect TScaledBitmapDrawer::GameVisibleArea() const
{
	return TRect(iScaledBitmapContext.Point(), iScaledBitmapContext.Size());
}

TSize TScaledBitmapDrawer::GameSize() const
{
	return iScaledBitmapContext.Size();
}

TPoint TScaledBitmapDrawer::GameOrigin() const
{
	return iScaledBitmapContext.Point();
}





void TGamePaner::ActivateL(CGameWindow* aGameWindow, CGameController* aGameController, TOrientation aOrientation)
{
	iGameWindow = aGameWindow;
	iGameController = aGameController;
	iOrientation = aOrientation;

	// wait for the first frame to see if it needs paning
	iGameWindow->AddGameScreenDrawObserverL(*this);
}

void TGamePaner::PostDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/)
{
	TRect	rect = iGameWindow->GameVisibleArea();
	TSize	size = iGameWindow->GameSize() - rect.Size();

	if (size.iWidth > 0)
	{
		iDeltaX = 2;
	}
	if (size.iHeight > 0)
	{
		iDeltaY = 2;
	}
	if (iDeltaX || iDeltaY)
	{
		// yup, needs paning, so make sure we get key events
		iGameController->AddGameControlInputHandlerL(*this);
	}

	iCurDelta = iGameWindow->GameOrigin();
	iMaxDelta = size.AsPoint();
	iGameWindow->RemoveGameScreenDrawObserver(*this);
}


void TGamePaner::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
	TPoint	delta(0, 0);

	if (aType == EEventKey)
	{
		switch (aKeyEvent.iCode)
		{
			case GAME_RIGHT_KEYCODE:
				delta.iX = (iCurDelta.iX + iDeltaX > iMaxDelta.iX) ? iMaxDelta.iX - iCurDelta.iX : iDeltaX;
				break;

			case GAME_LEFT_KEYCODE:
				delta.iX = (iCurDelta.iX < iDeltaX) ? iCurDelta.iX*-1 : iDeltaX*-1;
				break;

			case GAME_UP_KEYCODE:
				if (delta.iY)
					delta.iY = (iCurDelta.iY < iDeltaY) ? iCurDelta.iY*-1 : iDeltaY*-1;
				break;

			case GAME_DOWN_KEYCODE:
				if (delta.iY)
					delta.iY = (iCurDelta.iY + iDeltaY > iMaxDelta.iY) ? iMaxDelta.iY - iCurDelta.iY : iDeltaY;
				break;
		}
	}
	if (delta.iX || delta.iY)
	{
		iCurDelta += delta;
		iGameWindow->SetGameOrigin(iGameWindow->GameOrigin() - delta);
	}
}

/***********************************
 *
 * CInfoMgr implementation
 *
 ***********************************/

CInfoMgr::CInfoMgr(CGameWindow& aGameWindow, CGameController& aGameController) : 
	iGameWindow(aGameWindow), iGameController(aGameController), iPosition(20, 0), iSize(50, 12),
	iFgColor(KRgbWhite), iBgColor(KRgbBlack)
{
}

CInfoMgr::~CInfoMgr()
{
	if (iFont)
		iBitmap->FbsBitmapDevice()->ReleaseFont(iFont);
	if (iBitmap)
		delete iBitmap;
}

CInfoMgr* CInfoMgr::NewL(CGameWindow& aGameWindow, CGameController& aGameController)
{
	CInfoMgr*	self = new(ELeave)CInfoMgr(aGameWindow, aGameController);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

void CInfoMgr::ConstructL()
{
	TTimeIntervalMicroSeconds32	period;
	UserHal::TickPeriod(period);
	iPeriod	= period.Int();

	iBitmap = CFbsGameScreenDevice::NewL(iSize, iGameWindow.GameScreenDevice()->DisplayMode());

	TFontSpec fontSpec(KFontName, 20);
	User::LeaveIfError(iBitmap->FbsBitmapDevice()->GetNearestFontInPixels(iFont, fontSpec));

	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->UseFont(iFont);

	iGameController.AddGameControlInputHandlerL(*this);
	iGameController.AddGameControlRawInputHandlerL(*this);	// MR ZOOM
}

void CInfoMgr::PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice)
{
	TPoint p = iGameWindow.GameOrigin() + iPosition;
	if (p.iX < 0)
		p.iX = 0;
	if (p.iY < 0)
		p.iY = 0;

	DrawBitmap();
	aDstGameScreenDevice->DirectDraw(p, *iBitmap, TRect(iSize));
	GameTick();
}


void CInfoMgr::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
	if (aType == EEventKey)
	{
		if (aKeyEvent.iCode == '#')
		{
			if (!iIsActive)
			{
				ActivateL();
			}
			else
			{
				Deactivate();
			}
		}
	}
}


void CInfoMgr::DrawBitmap()
{
	TBuf<32>	text;

	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->SetBrushColor(iBgColor);
	gc->SetBrushStyle(CGraphicsContext::ENullBrush);
	gc->SetPenColor(iFgColor);
	gc->SetPenSize(TSize(1,1));
	gc->Clear();

	text.Format(_L("%d fps"), iVideoFps);
					
	// draw text
	TInt baselineOffset=(iSize.iHeight + iFont->AscentInPixels()) / 2;
	gc->DrawText(text, TRect(iSize), baselineOffset, CGraphicsContext::ECenter);
}

void CInfoMgr::ActivateL()
{
	iIsActive = ETrue;
	iVideoFps = 0;
	iFrames = 0;
	iGameWindow.AddGameScreenDrawObserverL(*this);
}

void CInfoMgr::Deactivate()
{
	iGameWindow.RemoveGameScreenDrawObserver(*this);
	iIsActive = EFalse;
}

void CInfoMgr::GameTick()
{
	iFrames++;
	if (iFrames > 50)
	{
		TInt ticks = User::TickCount();
		TInt diff = ticks - iPrevTicks;
		iPrevTicks = ticks;

		TInt time = diff * iPeriod;
		if (!time)
			time = 1;
		iVideoFps = (iFrames * KOneSecondInMicroSeconds) / time;
		iFrames = 0;
	}
}

/*************************************************
 *
 * CSmartPenInputHandler implementation
 *
 *************************************************/

CSmartPenInputHandler::CSmartPenInputHandler(CGameWindow& aGameWindow, CGameController& aGameController)
	: iGameWindow(aGameWindow), iGameController(aGameController), iFramesRepeatThreshHold(1), iFramesMoveThreshHold(1),
	iAutoX(ETrue), iAutoY(ETrue)
{
}


CSmartPenInputHandler::~CSmartPenInputHandler()
{
}

void CSmartPenInputHandler::ConstructL()
{
	iGameController.AddGameControlInputHandlerL(*this);
	iGameWindow.AddGameScreenDrawObserverL(*this);
	iGameWindow.AddScreenOrientationObserverL(*this);
}


void CSmartPenInputHandler::HandlePointerEvent(const TPointerEvent& aPointerEvent)
{
	if (aPointerEvent.iType == TPointerEvent::EButton1Down)
		PenDownEvent(aPointerEvent.iPosition);
	else if (aPointerEvent.iType == TPointerEvent::EButton1Up)
		PenUpEvent(aPointerEvent.iPosition);
	else if (aPointerEvent.iType == TPointerEvent::EDrag)
		PenMoveEvent(aPointerEvent.iPosition);
}

void CSmartPenInputHandler::PostDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/)
{
	if (iPenDown)
	{
			/*
			 * The first threashold is low so that we start moving relatively quick.
			 */
		if (++iFrameCount > iFramesThreshHold)
		{
			iFrameCount = 0;
			SimulateAutoMove();
			if (iEventCount == 2)
			{
				iFramesThreshHold = iFramesRepeatThreshHold;
			}
		}
	}
	else if (iEventCount)
	{
			/*
			 * Generate the up event for a fast pen stroke. For such a stroke
			 * we have generated the key down event on the pen-up. We have waited
			 * a couple of frames to generate the key-up event.
			 */
		if (++iFrameCount > iFramesThreshHold)
		{
			iFrameCount = 0;
			PenUpEvent(iLastPenPosition);
		}
	}
}

void CSmartPenInputHandler::SimulateAutoMove()
{
#ifdef PROJECT_EDOOM
	if (1)
#else
	if (iLastDelta.iX || iLastDelta.iY)
#endif
	{
		if (iEventCount == 0)
			HandleDownEvent();
		HandleMoveEvent(iLastDelta.iX, iLastDelta.iY);
		iEventCount++;
		if (!iAutoY)
			iLastDelta.iY = 0;
		if (!iAutoX)
			iLastDelta.iX = 0;
	}
}


void CSmartPenInputHandler::PenDownEvent(const TPoint& aPosition)
{
//	RDebug::Print(_L("CSmartPenInputHandler::PenDownEvent"));

		/*
		 * if iEventCount !=0 then we got a new down event when were handling
		 * a special fast short moved pen stroke. For these strokes we will
		 * wait a couple of frames after the pen-up event.
		 */
	if (iEventCount)
		PenUpEvent(iLastPenPosition);

	iDownPenPosition = iLastPenPosition = aPosition;
	iLastDelta = TPoint(0,0);
	iPenDown = ETrue;

	iFramesThreshHold = iFramesMoveThreshHold;
	iEventCount = iFrameCount = 0;

	HandleDownEvent();
}


void CSmartPenInputHandler::PenUpEvent(const TPoint& aPosition)
{
//	RDebug::Print(_L("CSmartPenInputHandler::PenUpEvent"));

	if (!iEventCount)
	{
			/*
			 * If we have not generated any events during the pen move
			 * then this maybe a very quick pen stroke. To allow for high
			 * precision movements we will consider all delta's to be a valid
			 * move (EFalse in PenMoveEvent). We set iFramesThreshHold so that
			 * we can generate the according up event a couple of frames later.
			 * Doom will ignore handle key-down and key-up events that happen 
			 * too short after one another.
			 */
		if (!iLastDelta.iX && !iLastDelta.iY)
			PenMoveEvent(aPosition, EFalse);
		HandleMoveEvent(iLastDelta.iX, iLastDelta.iY);
		iFramesThreshHold = iFramesMoveThreshHold;
	}
	else
	{
		HandleUpEvent();
		iFramesThreshHold = iEventCount = iLastDelta.iX = iLastDelta.iY = 0;
	}
	iPenDown = EFalse;
}


TPoint CSmartPenInputHandler::TransformDeltaForOrientation(const TPoint& aPoint)
{
	switch (iOrientation)
	{
		case EOrientationRotLeft:
			return TPoint(aPoint.iY, aPoint.iX * -1);
		case EOrientationRotRight:
			return TPoint(aPoint.iY * -1, aPoint.iX);
		case EOrientationRotUpsideDown:
			return TPoint(aPoint.iX * -1, aPoint.iY * -1);
	}
	return aPoint;
}


void CSmartPenInputHandler::PenMoveEvent(const TPoint& aPosition, TBool aStrict)
{
	TPoint	deltaRel = aPosition - iLastPenPosition;

	if (aStrict)
	{
			/*
			 * Filter out jitter
			 */
		if ((deltaRel.iX < 10) && (deltaRel.iX > -10))
			deltaRel.iX = 0;
		if ((deltaRel.iY < 10) && (deltaRel.iY > -10))
			deltaRel.iY = 0;
	}

	if (deltaRel.iX || deltaRel.iY)
	{
		deltaRel = TransformDeltaForOrientation(deltaRel);
		iLastPenPosition = aPosition;
		iLastDelta = deltaRel;
//		RDebug::Print(_L("CSmartPenInputHandler::PenMoveEvent (%d, %d), dx=%d, dy=%d"), aPosition.iX, aPosition.iY, delta.iX, delta.iY);
	}
}


/*************************************************
 *
 * CSmartPenToKeyInputHandler implementation
 *
 *************************************************/

CSmartPenToKeyInputHandler::CSmartPenToKeyInputHandler(CGameWindow& aGameWindow, CGameController& aGameController)
	: CSmartPenInputHandler(aGameWindow, aGameController)
{
}

void CSmartPenToKeyInputHandler::ConstructL()
{
	CSmartPenInputHandler::ConstructL();
}


CSmartPenToKeyInputHandler* CSmartPenToKeyInputHandler::NewL(CGameWindow& aGameWindow, CGameController& aGameController)
{
	CSmartPenToKeyInputHandler*		self = new(ELeave)CSmartPenToKeyInputHandler(aGameWindow, aGameController);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

void CSmartPenToKeyInputHandler::HandleDownEvent()
{
	iHorMove = iVertMove = 0;
	iIsMovingVert = EFalse;
	iLastPenPosition1 = TPoint(0,0);
}


void CSmartPenToKeyInputHandler::HandleHorMoveEvent(TInt aNewHorMove)
{
	if (aNewHorMove != iHorMove)
	{
		if (iHorMove)
		{
			if (iHorMove==-1)
			{
				InjectkeyEvent(EEventKeyUp, GAME_LEFT_SCANCODE);
				iLeftCount = 0;
			}
			else
			{
				InjectkeyEvent(EEventKeyUp, GAME_RIGHT_SCANCODE);
				iRightCount = 0;
			}
		}

		if (aNewHorMove < 0)
		{
			InjectkeyEvent(EEventKeyDown, GAME_LEFT_SCANCODE);
			InjectkeyEvent(EEventKey, GAME_LEFT_KEYCODE);
		} 
		else if (aNewHorMove>0)
		{
			InjectkeyEvent(EEventKeyDown, GAME_RIGHT_SCANCODE);
			InjectkeyEvent(EEventKey, GAME_RIGHT_KEYCODE);
		}
	}
	else
	{
		if (aNewHorMove < 0)
			InjectkeyEvent(EEventKey, GAME_LEFT_KEYCODE, ++iLeftCount);
		else if (aNewHorMove>0)
			InjectkeyEvent(EEventKey, GAME_RIGHT_KEYCODE, ++iRightCount);
	}


	iHorMove = aNewHorMove;
}

void CSmartPenToKeyInputHandler::HandleVertMoveEvent(TInt aNewVertMove)
{
	if (aNewVertMove != iVertMove)
	{
		if (iVertMove)
		{
			if (iVertMove==-1)
			{
				InjectkeyEvent(EEventKeyUp, GAME_UP_SCANCODE);
				iUpCount = 0;
			}
			else
			{
				InjectkeyEvent(EEventKeyUp, GAME_DOWN_SCANCODE);
				iDownCount = 0;
			}
		}

		if (aNewVertMove < 0)
		{
			InjectkeyEvent(EEventKeyDown, GAME_UP_SCANCODE);
			InjectkeyEvent(EEventKey, GAME_UP_KEYCODE);
		} 
		else if (aNewVertMove>0)
		{
			InjectkeyEvent(EEventKeyDown, GAME_DOWN_SCANCODE);
			InjectkeyEvent(EEventKey, GAME_DOWN_KEYCODE);
		}
	}
	else
	{
		if (aNewVertMove < 0)
			InjectkeyEvent(EEventKey, GAME_UP_KEYCODE, ++iUpCount);
		else if (aNewVertMove>0)
			InjectkeyEvent(EEventKey, GAME_DOWN_KEYCODE, ++iDownCount);
	}

	iVertMove = aNewVertMove;
}


void CSmartPenToKeyInputHandler::HandleMoveEvent(TInt aX, TInt aY)
{
//	RDebug::Print(HERE("GenerateEvent: ax=%d, ay=%d"), aX, aY);
	TInt absX = abs(aX);
	TInt absY = abs(aY);

	if (!iCanDo_X_And_Y_AtTheSameTime)
	{
			/*
			 * If we want to move in one direction only, ignore either aX or aY
			 */
		if (absX > absY)
			aY = 0;
		else 
			aX = 0;
	}

	TInt newHorMove = (aX < 0)? -1 : (aX>0) ? 1 : 0;
	TInt newVertMove = (aY < 0)? -1 : (aY>0) ? 1 : 0;

	if (iEnableAutoForward)
	{
//		TPoint deltaSinceLastVertMove = iLastPenPosition1 - iLastPenPosition;
//		deltaSinceLastVertMove = this->TransformDeltaForOrientation(deltaSinceLastVertMove);
//		if (newVertMove!=0)
//			iLastPenPosition1 = iLastPenPosition;

		iIsMovingVert = (iVertMove != 0);
//		iAutoX = iIsMovingVert ? EFalse : ETrue;
		if (iIsMovingVert && (newVertMove == 0))
			newVertMove = iVertMove;

		if (iHorMove && newHorMove && (iHorMove != newHorMove))
		{
			newHorMove = 0;
			this->iLastDelta.iX = 0;
		}
//		if (newHorMove)
//			InjectkeyEvent(EEventKey, (newHorMove > 0) ? 1 : 2);
	}

	if (absX > absY)
	{
		HandleHorMoveEvent(newHorMove);
		HandleVertMoveEvent(newVertMove);
	}
	else
	{
		HandleVertMoveEvent(newVertMove);
		HandleHorMoveEvent(newHorMove);
	}

//	RDebug::Print(HERE("GenerateEvent: hor=%d, vert=%d\n"), iHorMove, iVertMove);
}


void CSmartPenToKeyInputHandler::HandleUpEvent()
{
	if (iHorMove)
		InjectkeyEvent(EEventKeyUp, (iHorMove==-1)? GAME_LEFT_SCANCODE : GAME_RIGHT_SCANCODE);
	if (iVertMove)
		InjectkeyEvent(EEventKeyUp, (iVertMove==-1)? GAME_UP_SCANCODE : GAME_DOWN_SCANCODE);
	iLeftCount = iRightCount = iUpCount = iDownCount = 0;
}


void CSmartPenToKeyInputHandler::InjectkeyEvent(TInt aEventType, TInt aKeyOrScanKey, TInt aRepeat)
{
	iGameController.InjectkeyEvent(aEventType, aKeyOrScanKey, aRepeat);
	iEventCount++;
}


/**********************************************
 *
 * CGameIntro
 *
 **********************************************/

CGameIntro::CGameIntro(CGameWindow& aGameWindow, CGameController& aGameController) : 
	iGameWindow(aGameWindow), iGameController(aGameController), iPosition(0, 0), iSize(200, 40),
	iFgColor(KRgbWhite), iBgColor(KRgbBlack), iIsActive(ETrue)
{
}

CGameIntro::~CGameIntro()
{
	Destruct();
}



CGameIntro* CGameIntro::NewL(CGameWindow& aGameWindow, CGameController& aGameController)
{
	CGameIntro*	self = new(ELeave)CGameIntro(aGameWindow, aGameController);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

void CGameIntro::ConstructL()
{
	iBitmap = CFbsGameScreenDevice::NewL(iSize, iGameWindow.GameScreenDevice()->DisplayMode());

	TFontSpec fontSpec(KFontName, 20);
	User::LeaveIfError(iBitmap->FbsBitmapDevice()->GetNearestFontInPixels(iFont, fontSpec));

	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->UseFont(iFont);

	iGameController.AddGameControlInputHandlerL(*this);
	iGameWindow.AddGameScreenDrawObserverL(*this);

	DrawBitmap();
	BitmapRecolor();
}

void CGameIntro::Destruct()
{
	if (iFont)
	{
		iBitmap->FbsBitmapDevice()->ReleaseFont(iFont);
		iFont = NULL;
	}
	if (iBitmap)
	{
		delete iBitmap;
		iBitmap = NULL;
	}
}


static TRgb sRgbTable[] = 
{
	TRgb(81, 81, 236),
	TRgb(78, 138, 239),
	TRgb(77, 171, 240),
	TRgb(77, 203, 240),
	TRgb(77, 240, 227),
	TRgb(75, 241, 175),
	TRgb(75, 241, 130),
	TRgb(79, 242, 74),
	TRgb(133, 243, 73),
	TRgb(188, 243, 73),
	TRgb(243, 243, 73),
	TRgb(244, 188, 72),
	TRgb(245, 132, 71),
	TRgb(245, 71, 71),
	TRgb(247, 70, 132),
	TRgb(248, 69, 194),
	TRgb(243, 69, 248),
	TRgb(181, 68, 249),
	TRgb(177, 67, 250),
	TRgb(70, 66, 251)
};

static TInt iRgbIndex = 0;
#define MAX_RGB_TABLE	(sizeof(sRgbTable)/sizeof(TRgb))

void CGameIntro::PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice)
{
	TPoint p = iGameWindow.GameOrigin() + iPosition;
	if (p.iX < 0)
		p.iX = 0;
	if (p.iY < 0)
		p.iY = 0;

	if (++iFrames == 6)
	{
		iFrames = 0;
		if (++iRgbIndex == MAX_RGB_TABLE)
			iRgbIndex =0;
		DrawBitmap();
		BitmapRecolor();
	}
	aDstGameScreenDevice->DirectDraw(p, *iBitmap, TRect(iSize));
}


TBool CGameIntro::WantDrawToScreen()
{
	return iIsActive;
}


void CGameIntro::HandleKeyEvent(TInt aType, const TKeyEvent& /*aKeyEvent*/)
{
	if (aType == EEventKey)
	{
//		iGameController.RemoveGameControlInputHandler(*this);
		iGameWindow.RemoveGameScreenDrawObserver(*this);
		iIsActive = EFalse;
//		Destruct();
	}
}

void CGameIntro::BitmapRecolor()
{
	TDisplayMode displayMode = iGameWindow.GameScreenDevice()->DisplayMode();
	TUint16		white = (TUint16) TGameBitmapUtil::GetColorValue(KRgbWhite.Color64K(), displayMode);
	TInt		index = iRgbIndex;
	TUint16		color;

	for (TInt h=0 ; h<iSize.iHeight ; h++)
	{
		TUint16*	p = (TUint16*) iBitmap->ScanLineAddress(h);

		TInt		index1 = index;
		TInt		count = 0;
		color = (TUint16) TGameBitmapUtil::GetColorValue(sRgbTable[index1], displayMode);
		for (TInt w=0 ; w<iSize.iWidth ; w++)
		{
			if (*p == white)
				*p = color;
			p++;

			if (++count == 16)
			{
				if (++index1 == MAX_RGB_TABLE)
					index1 = 0;
				count = 0;
				color = (TUint16) TGameBitmapUtil::GetColorValue(sRgbTable[index1], displayMode);
			}
		}
		if (++index == MAX_RGB_TABLE)
		{
			index = 0;
		}
	}
}


void CGameIntro::DrawBitmap()
{
	TBuf<32>	text;

	CFbsBitGc*	gc = iBitmap->FbsBitGc();
	gc->SetBrushColor(iBgColor);
	gc->SetBrushStyle(CGraphicsContext::ENullBrush);
	gc->SetPenColor(iFgColor);
	gc->SetPenSize(TSize(1,1));
	gc->Clear();


	// draw text
	TRect	rect(TSize(iSize.iWidth, 10));
	text.Format(_L("Symbian OS port"));
	TInt baselineOffset=(rect.Size().iHeight + iFont->AscentInPixels()) / 2;
	gc->DrawText(text, rect, baselineOffset, CGraphicsContext::ECenter);

	text.Format(_L("by"));
	rect.Move(0, 12);
	baselineOffset=(rect.Size().iHeight + iFont->AscentInPixels()) / 2;
	gc->DrawText(text, rect, baselineOffset, CGraphicsContext::ECenter);
	
	text.Format(_L("Peter van Sebille"));
	rect.Move(0, 12);
	baselineOffset=(rect.Size().iHeight + iFont->AscentInPixels()) / 2;
	gc->DrawText(text, rect, baselineOffset, CGraphicsContext::ECenter);
}


/******************************************
 *
 * CGameDirectScreenAccess
 *
 ******************************************/

CGameDirectScreenAccess::~CGameDirectScreenAccess()
{
	if (iRequestStatus == KRequestPending)
		iDirectScreenAccess.Cancel();
	if (iDirectScreenAccess.WsHandle() != -1)
		iDirectScreenAccess.Close();

	if (iSemaphore.Handle() != -1)
		iSemaphore.Close();
}

CGameDirectScreenAccess* CGameDirectScreenAccess::NewL(CGameWindow& aGameWindow)
{
	CGameDirectScreenAccess*	self = new(ELeave)CGameDirectScreenAccess(aGameWindow);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

CGameDirectScreenAccess::CGameDirectScreenAccess(CGameWindow& aGameWindow) :
	iGameWindow(aGameWindow)
{
}


void CGameDirectScreenAccess::ConstructL()
{
	iDirectScreenAccess = RDirectScreenAccess(iGameWindow.iWsSession);
	User::LeaveIfError(iDirectScreenAccess.Construct());
	TBuf<32>	semaphoreName;
	semaphoreName.Format(_L("%S_SEM"), &iGameWindow.iGameName);
	User::LeaveIfError(iSemaphore.CreateGlobal(semaphoreName, 0));

	ClaimScreenL();
}


void CGameDirectScreenAccess::ClaimScreenL()
{
	iRequestStatus = KRequestPending;

	RRegion*	region;
	User::LeaveIfError(iDirectScreenAccess.Request(region, iRequestStatus, iGameWindow.iWsWindow));
	TBool	isFullScreen = region->IsContainedBy(TRect(iGameWindow.iWsWindow.Position(), iGameWindow.iWsWindow.Size()));
	if (!isFullScreen)
		User::Leave(KErrInUse);
}


#if 0
void DumpWindowGroups(RWsSession& aWsSession)
{
	TInt			id = 0;
	TInt			prio = 0;
	TFullName		threadName;
	TFullName		wgName;

	RDebug::Print(_L("************************"));
	while (id >= 0)
	{
		id = aWsSession.FindWindowGroupIdentifier(id , _L("*"));
		if (id > 0)
		{
			TThreadId threadId;
			if (aWsSession.GetWindowGroupClientThreadId(id , threadId) == KErrNone)
			{
				threadName = _L("");
				RThread	thread;
				if (thread.Open(threadId) == KErrNone)
				{
					threadName = thread.Name();
					thread.Close();
				}

				wgName = _L("");
				aWsSession.GetWindowGroupNameFromIdentifier(id, wgName);
				prio = aWsSession.GetWindowGroupOrdinalPriority(id);
			}
			RDebug::Print(_L("id=%d, thread=%S, wg=%S, prio=%d"), id, &threadName, &wgName, prio);
		}
	}
	RDebug::Print(_L("************************"));
}
#endif

void CGameDirectScreenAccess::PauseGameL()
{
	iDirectScreenAccess.Completed();

	User::After(TTimeIntervalMicroSeconds32(1000000));
	iGameWindow.Hide();

//	DumpWindowGroups(iGameWindow.iWsSession);
	iSemaphore.Wait();
	ResumeGameL();
}

void CGameDirectScreenAccess::ResumeGameL()
{
	iGameWindow.Show();
	ClaimScreenL();
	iGameWindow.ClearScreen();
}

void CGameDirectScreenAccess::PreDrawFrameToScreen(CGameScreenDevice* /*aSrcGameScreenDevice*/)
{
	if (iRequestStatus != KRequestPending)
		PauseGameL();
}


