/*******************************************************************
 *
 *	File:		Video.h
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#ifndef __VIDEO_H
#define __VIDEO_H


#include <e32base.h>
#include <w32std.h>

#include "gameengine.h"
#include "navigation.h"

const TInt KMaxCapturedButtons = 5;

enum TOrientation
{
	EOrientationNormal,
	EOrientationRotLeft,
	EOrientationRotRight,
	EOrientationRotUpsideDown,
};

const TInt KVideoModes = EColor4K +1;
class CGameScreenDevice;
class TScaledBitmapContext
{
public:
	TScaledBitmapContext();
	TScaledBitmapContext(const CGameScreenDevice& aDstGameScreenDevice, const CGameScreenDevice& aSrcGameScreenDevice, const TSize& aBitmapSize, TBool aCenter);
	TInt SkipX() const {return iSkipX;}
	TInt SkipY() const {return iSkipY;}
	TPoint	Point() const {return iPoint;}
	TSize	Size() const {return iSize;}
	const CGameScreenDevice& SrcGameScreenDevice() const {return *iSrcGameScreenDevice;}
	TUint*	ScanLine() const {return iScanLine;}
	TInt	ScanLineBytes() const {return iScanLineBytes;}
	TBool	HasSetup()const {return iSrcGameScreenDevice!=NULL;}

protected:
	TInt	iSkipX;
	TInt	iSkipY;
	TPoint	iPoint;
	TSize	iSize;
	const CGameScreenDevice* iSrcGameScreenDevice;
	mutable TUint	iScanLine[1024];
	TInt	iScanLineBytes;
};

class CExtraFbsBitmap : public CFbsBitmap
{
public:
	void GameBitmapLockHeap() {CFbsBitmap::LockHeap();}
	CBitwiseBitmap* BitmapAdressAddress() const {return iAddressPointer;}
};


class CGamePalette : public CBase
{
public:
	~CGamePalette();
	static CGamePalette* NewL(TDisplayMode aDisplayMode);

	// palette operations
	TUint*				PaletteBuffer() const {return (TUint*) iPalette;}
	TInt				AllocateEntry(const TRgb& aColor);
	TDisplayMode		DisplayMode() const {return iDisplayMode;}

protected:
	CGamePalette(TDisplayMode aDisplayMode);
	void ConstructL();

	TUint				iPalette[256];
	TDisplayMode		iDisplayMode;
	TInt				iPaletteLastIndexUsed;
	TInt				iPaletteMaxIndex;
};

class CGameScreenDevice : public CBase
{
public:
	~CGameScreenDevice();

	// getters
	virtual TUint8*		ScanLineAddress(TInt aLine) const = 0;
	TInt				BytesPerLine() const {return iBytesPerLine;}
	TSize				Size() const {return iSize;}
	TDisplayMode		DisplayMode() const {return iDisplayMode;}

	// draw operations
	void	Clear();
	void	DirectDraw(const TPoint& aPoint, const CGameScreenDevice& aSrcDevice, const TRect& aRect);
	void	DirectDoubleDraw(const TPoint& aPoint, const CGameScreenDevice& aSrcDevice, const TRect& aRect);
	void	ScaledDrawBitmap(const TScaledBitmapContext& aScaledBitmapContext);
	void	AlphaBlend(const TPoint& aPoint, const CGameScreenDevice& aSrcDevice, const TRect& aRect, TInt8 aAlpha);

	void DirectDraw_16bpp_16bppRotLeft(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);
	void DirectDraw_16bpp_16bppRotRight(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);

	// palette operations
	CGamePalette*		Palette() const {return iPalette;}
	void				SetPalette(CGamePalette* aGamePalette) {iPalette = aGamePalette;}

protected:
	CGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode);

	void	DirectDraw_8bpp_8bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);
	void	DirectDraw_12or16bpp_8bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);
	void	DirectDraw_12or16bpp_12or16bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);
	void	DirectDraw_Unsupported(const TPoint& /*aDstOrigin*/, const CGameScreenDevice& /*aSrcDevice*/, const TPoint& /*aSrcOrigin*/, const TSize& /*aSize*/){ASSERT(EFalse);}

	void	DoDirectDoubleDraw_Unsupported(const TPoint& /*aDstOrigin*/, const CGameScreenDevice& /*aSrcDevice*/, const TPoint& /*aSrcOrigin*/, const TSize& /*aSize*/){ASSERT(EFalse);}
	void	DoDirectDoubleDraw_12bpp_8bpp(const TPoint& /*aDstOrigin*/, const CGameScreenDevice& /*aSrcDevice*/, const TPoint& /*aSrcOrigin*/, const TSize& /*aSize*/){ASSERT(EFalse);}

	void	ScaledScanLineDraw_Unsupported(const TScaledBitmapContext& /*aScaledBitmapContext*/, TUint* /*aSrcLine*/, TBool /*aMerge*/){ASSERT(EFalse);}
	void	ScaledScanLineDraw_12or16bpp_8bpp(const TScaledBitmapContext& aScaledBitmapContext, TUint* aSrcLine, TBool aMerge);

	void	AlphaBlend_12bpp_12bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize, TUint8 aAlpha);
	void	AlphaBlend_16bpp_16bpp(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize, TUint8 aAlpha);
	void	AlphaBlend_Unsupported(const TPoint& /*aDstOrigin*/, const CGameScreenDevice& /*aSrcDevice*/, const TPoint& /*aSrcOrigin*/, const TSize& /*aSize*/, TUint8 /*aAlpha*/){ASSERT(EFalse);}

	TInt				iBytesPerLine;
	TSize				iSize;
	TDisplayMode		iDisplayMode;
	CGamePalette*		iPalette;

protected:
	typedef void (CGameScreenDevice::*TDirectDrawer)(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize);
	static const TDirectDrawer KDirectDrawers12or16bpp[KVideoModes];
	static const TDirectDrawer KDirectDrawers8bpp[KVideoModes];

	typedef void (CGameScreenDevice::*TAlphaBlender)(const TPoint& aDstOrigin, const CGameScreenDevice& aSrcDevice, const TPoint& aSrcOrigin, const TSize& aSize, TUint8 aAlpha);
	static const TAlphaBlender KAlphaBlenders12bpp[KVideoModes];
	static const TAlphaBlender KAlphaBlenders16bpp[KVideoModes];

	typedef void (CGameScreenDevice::*TScaledScanLineDrawer)(const TScaledBitmapContext& aScaledBitmapContext, TUint* aSrcLine, TBool aMerge);
	static const TScaledScanLineDrawer KScaledScanLineDrawers12or16bpp[KVideoModes];
};


class CFbsGameScreenDevice : public CGameScreenDevice
{
public:
	~CFbsGameScreenDevice();
	static CFbsGameScreenDevice* NewL(const TSize& aSize, TDisplayMode aDisplayMode);
	virtual TUint8*		ScanLineAddress(TInt aLine) const;

	CFbsBitmap*			FbsBitmap(){return iBitmap;}
	CFbsBitGc*			FbsBitGc(){return iBitmapContext;}
	CFbsBitmapDevice*	FbsBitmapDevice(){return iBitmapDevice;}

protected:
	CFbsGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode);
	void ConstructL();

	CFbsBitGc*				iBitmapContext;
	CExtraFbsBitmap*		iBitmap;
	CFbsBitmapDevice*		iBitmapDevice;

	TUint32*				iDataAddress;
	CBitwiseBitmap*			iBitwiseBitmap;
};



class CHwFbGameScreenDevice : public CGameScreenDevice
{
public:
	static CHwFbGameScreenDevice* NewL(const TSize& aWindowSize, TDisplayMode aDisplayMode);

	virtual TUint8*		ScanLineAddress(TInt aLine) const;
protected:
	CHwFbGameScreenDevice(const TSize& aWindowSize, TDisplayMode aDisplayMode);
	void ConstructL(const TScreenInfoV01& aScreenInfo);
	TUint8*				iFrameBuffer;
};



class CMemGameScreenDevice : public CGameScreenDevice
{
public:
	~CMemGameScreenDevice();
	static CMemGameScreenDevice* NewL(const TSize& aSize, TDisplayMode aDisplayMode, TUint8* aMemBuffer=NULL);

	virtual TUint8*		ScanLineAddress(TInt aLine) const;
protected:
	CMemGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode, TUint8* aMemBuffer=NULL);
	void ConstructL();

	TUint8*				iMemBuffer;
	TBool				iOwnsMemBuf;
};


/*****************************************
 *
 * BITMAP SUPPORT
 *
 *****************************************/


class TGameBitmapUtil
{
public:
	static TDisplayMode GetDisplayMode(TInt aBpp, TBool aColor = ETrue);
	static TInt			GetBpp(TDisplayMode aDisplayMode);
	static TInt			GetColorDepth(TDisplayMode aDisplayMode);
	static TUint8*		PixelAddress(const CGameScreenDevice* aSrc, const TPoint& aPoint);
	static TRect		VisibleIntersection(const TSize& aScreenSize, const TPoint& aOrigin, const TSize& aBitmapSize, const TRect& aBitmapRect);
	static TUint		GetColorValue(const TRgb& aColor, const TDisplayMode& aDisplayMode);
};


template<class T>
void RemoveFromPointerArrayList(CArrayPtrFlat<T>& aList, T* aT)
{
	TInt	count = aList.Count();
	for (TInt i=count-1 ; i>=0 ; i--)
		if ((aList)[i] == aT)
			aList.Delete(i);
}


class CGameWindow;
class CVirtualKeyboard;
class CGameController;
class CInfoMgr;
class CSmartPenInputHandler;
class CSmartPenToKeyInputHandler;
class CGameIntro;
class CGameDirectScreenAccess;


class MGameScreenDrawer
{
public:
	virtual void	DrawFrame(CGameScreenDevice* aDstGameScreenDevice, const CGameScreenDevice* aSrcGameScreenDevice) = 0;
	virtual TRect	GameVisibleArea() const = 0;
	virtual TSize	GameSize() const = 0;
	virtual TPoint	GameOrigin() const = 0;
	virtual void	SetGameOrigin(const TPoint& aOrigin) = 0;
};


class MGameScreenDrawObserver
{
public:
		/*
		 * WantDrawToScreen indicates whether the observer will modify the screen
		 * in PostDrawFrameToScreen. If so, then the game frame will be drawn to
		 * a memory screen device. The same memory screen device will then be
		 * passed in to PreDrawFrameToScreen and PostDrawFrameToScreen for 
		 * the observers to modify. Afterwards the memory screen device is blitted
		 * to screen.
		 */
	virtual TBool	WantDrawToScreen() {return EFalse;}
	virtual void	PreDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/){};
	virtual void	PostDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/){};
};



class MGameControlInputHandler
{
public:
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent) = 0;
	virtual void HandlePointerEvent(const TPointerEvent& aPointerEvent) = 0;
};


class MGameControlRawInputHandler
{
public:
	virtual TBool ConsumeRawKeyEvent(TInt aType, const TKeyEvent& aKeyEvent) = 0;
	virtual TBool ConsumeRawPointerEvent(const TPointerEvent& aPointerEvent) = 0;
};

class MScreenOrientationObserver
{
public:
	virtual void ScreenOrientationChanged(TOrientation aOrientation) = 0;
};

/**********************************************
 *
 * TDirectBitmapDrawer
 *
 **********************************************/

class TDirectBitmapDrawer : public MGameScreenDrawer
{
public:
	virtual void	DrawFrame(CGameScreenDevice* aDstGameScreenDevice, const CGameScreenDevice* aSrcGameScreenDevice);
	virtual TRect	GameVisibleArea() const;
	virtual TSize	GameSize() const;
	virtual TPoint	GameOrigin() const;
	virtual void	SetGameOrigin(const TPoint& aOrigin);

protected:
	TPoint		iGameOrigin;
	TRect		iGameSize;
	TRect		iGameVisibleArea;
	TBool		iSetup;
};


/**********************************************
 *
 * TScaledBitmapDrawer
 *
 **********************************************/

class TScaledBitmapDrawer : public MGameScreenDrawer
{
public:
	virtual void	DrawFrame(CGameScreenDevice* aDstGameScreenDevice, const CGameScreenDevice* aSrcGameScreenDevice);
	virtual TRect	GameVisibleArea() const;
	virtual TSize	GameSize() const;
	virtual TPoint	GameOrigin() const;
	virtual void	SetGameOrigin(const TPoint& /*aOrigin*/){};

protected:
	TScaledBitmapContext	iScaledBitmapContext;
};

/********************************************
 *
 * TGamePaner
 *
 ********************************************/

class TGamePaner : MGameControlInputHandler, MGameScreenDrawObserver
{
public:
	void ActivateL(CGameWindow* aGameWindow, CGameController* aGameController, TOrientation aOrientation);

public:		// implements MGameScreenDrawObserver
	virtual void PostDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/);

public:		// implements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
	virtual void HandlePointerEvent(const TPointerEvent& /*aPointerEvent*/){};

protected:
	TInt				iDeltaX;
	TInt				iDeltaY;
	TPoint				iMaxDelta;
	TPoint				iCurDelta;
	CGameWindow*		iGameWindow;
	CGameController*	iGameController;
	TOrientation		iOrientation;
};





class CGameWindow : public CBase
{
public:
	~CGameWindow();
	static CGameWindow* NewL(const TDesC& aGameName);

	CGameScreenDevice*	GameScreenDevice(){return iGameScreenDevice;}
	void GameFrameTick(CGameScreenDevice* aSrcGameScreenDevice);

	CGameController& GameController() const {return *iGameController;}
	CVirtualKeyboard& VirtualKeyboard() const {return *iVirtualKeyboard;}

	void ClearScreen();

	// area on screen "covered" by the game; this is subject to scaling etc
	TRect	GameVisibleArea() const {return iGameScreenDrawer->GameVisibleArea();}
	TSize	GameSize() const {return iGameScreenDrawer->GameSize();}
	TPoint	GameOrigin() const {return iGameScreenDrawer->GameOrigin();}
	void	SetGameOrigin(const TPoint& aOrigin) {iGameScreenDrawer->SetGameOrigin(aOrigin);}

	void AddScreenOrientationObserverL(MScreenOrientationObserver& aObserver);
	void SetScreenOrientation(TOrientation aOrientation);

	void AddGameScreenDrawObserverL(MGameScreenDrawObserver& aObserver){iGameScreenDrawObservers->AppendL(&aObserver);}
	void RemoveGameScreenDrawObserver(MGameScreenDrawObserver& aObserver){RemoveFromPointerArrayList<MGameScreenDrawObserver>(*iGameScreenDrawObservers, &aObserver);}
	void SetGameScreenDrawer(MGameScreenDrawer* aGameScreenDrawer);

	void ConstructSmartPenToKeyInputHandlerL();
	void CanDo_X_And_Y_AtTheSameTime(TBool aCanDo);

	void Hide();
	void Show();

	void CaptureKeyEvents();
	void ReleaseKeyEvents();
	TInt32	iCapturedButtons[KMaxCapturedButtons];

	RWsSession& WsSession() {return iWsSession;}

protected:
	CGameWindow();
	void ConstructL(const TDesC& aGameName);

	MGameScreenDrawer*		iGameScreenDrawer;
	TScaledBitmapDrawer		iScaledBitmapDrawer;
	TDirectBitmapDrawer		iDirectBitmapDrawer;
	TGamePaner				iGamePaner;

		/*
		 * Our game device
		 * For target builds it's a CHwFbGameScreenDevice object, for WINS 
		 * a CFbsGameScreenDevice object
		 */
	void ConstructScreenDeviceL();
	void DestructScreenDevice();
	CGameScreenDevice*	iGameScreenDevice;

		/*
		 * Window Server stuff
		 */
	void WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice);
	void WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice, const TRect& aRect);
	void ConstructWsResourcesL();
	void DestructWsResources();
	CGameDirectScreenAccess*	iGameDirectScreenAccess;
	RWsSession					iWsSession;
	RWindowGroup				iWsWindowGroup;
	RWindow						iWsWindow;
	CWsScreenDevice*			iWsScreen;
	CWindowGc*					iWindowGc;
	TBool						iUseWServ;

	void ConstructGameControllerL();
	void DestructGameController();
	CGameController*		iGameController;

		/*
		 * Virtual keyboard stuff
		 */
	void ConstructVirtualKeyboardL();
	void DestructVirtualKeyboard();
	CVirtualKeyboard*	iVirtualKeyboard;

	CInfoMgr*						iInfoMgr;
	CSmartPenToKeyInputHandler*		iSmartPenToKeyInputHandler;
	CGameIntro*						iGameIntro;

	CArrayPtrFlat<MGameScreenDrawObserver>*	iGameScreenDrawObservers;
	CArrayPtrFlat<MScreenOrientationObserver>*	iScreenOrientationObservers;

	TOrientation iOrientation;

	TBuf<32>	iGameName;

	friend		CGameDirectScreenAccess;

	CMemGameScreenDevice*	iMemoryGameScreenDevice;
	TBool					iUsedMemoryScreenDevice;
	TBool					iWasPaused;
};



class CGameController : public CBase, public MGameScreenDrawObserver, public MScreenOrientationObserver
{
public:
	~CGameController();
	static CGameController* NewL(RWsSession& aWsSession);

	void AddGameControlRawInputHandlerL(MGameControlRawInputHandler& aControlImputConsumser){iGameControlRawInputHandlers->AppendL(&aControlImputConsumser);}
	void AddGameControlInputHandlerL(MGameControlInputHandler& aGameControlObserver){iGameControlInputHandlers->AppendL(&aGameControlObserver);}
	void RemoveGameControlRawInputHandler(MGameControlRawInputHandler& aControlImputConsumser){RemoveFromPointerArrayList<MGameControlRawInputHandler>(*iGameControlRawInputHandlers, &aControlImputConsumser);}
	void RemoveGameControlInputHandler(MGameControlInputHandler& aGameControlObserver){RemoveFromPointerArrayList<MGameControlInputHandler>(*iGameControlInputHandlers, &aGameControlObserver);}

	void PauseGame(TBool aDoPause);
	void InjectkeyEvent(TInt aEventType, TInt aKeyOrScanKey, TInt aRepeat=0);

	void SetGameKeymapping(TGameScanCodeTable& aGameScanCodeTable, TGameKeyCodeTable& aGameKeyCodeTable);

public:			// implements MGameScreenDrawObserver
	virtual void PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice);

public:			// implements MScreenOrientationObserver
	virtual void ScreenOrientationChanged(TOrientation aOrientation);

protected:
	CGameController(RWsSession& aWsSession) : iWsSession(aWsSession){}
	void ConstructL();

	void MapToGameEvents(TInt aType, TKeyEvent& aKeyEvent);
	TGameScanCodeTable*		iGameScanCodeTable;
	TGameKeyCodeTable*		iGameKeyCodeTable;

	TBool				iIsPaused;

		/*
		 * Raw pointer events
		 */
	void HandlePointerEvent(const TPointerEvent& aPointerEvent);
	void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);

	TKeyEvent TranslateEventForOrientation(TInt aType, const TKeyEvent& aKeyEvent);
	TOrientation	iOrientation;

		/*
		 * Window Server stuff
		 */
	void HandleWsEvent(const TWsEvent& aWsEvent);
	void HandleRedrawEvent(const TWsRedrawEvent& iRedrawEvent);
	void StartWServEvents();
	void PollWServEvents();
	void CancelWServEvents();

	RWsSession&			iWsSession;
	TRequestStatus		iWsEventStatus;
	TRequestStatus		iRedrawEventStatus;
	TWsEvent			iWsEvent;
	TWsRedrawEvent		iRedrawEvent;

	CArrayPtrFlat<MGameControlInputHandler>*	iGameControlInputHandlers;
	CArrayPtrFlat<MGameControlRawInputHandler>*	iGameControlRawInputHandlers;
};





/***********************************
 *
 * Virtual keyboard
 *
 ***********************************/


struct TVirtualKey
{
	TInt	iWidth;				// width in pixels
	TInt	iScanCode;
	char*	iText;				// text displayed in normal mode
	TInt	iKeyCode;			// generated key code
	char*	iShiftText;			// text displayed in when the shift key is pressed
	TInt	iShiftKeyCode;		// generated key code when the shift key is pressed
};

struct TVirtualButton
{
	TInt	iTlX;
	TInt	iTlY;
	TInt	iBrX;
	TInt	iBrY;
	TInt	iScanCode;
	TInt	iKeyCode;			// generated key code
};

class CVirtualKeyboard :	public CBase, 
							public MGameScreenDrawObserver, 
							public MGameControlRawInputHandler,
							public MGameControlInputHandler,
							public MScreenOrientationObserver
{
public:
	~CVirtualKeyboard();
	static CVirtualKeyboard* NewL(CGameWindow& aGameWindow, CGameController& aGameController);
	
	void SetPosition(const TPoint& aPosition);
	void SetActiveRect(const TRect& aRect){iActiveRect=aRect;}

	void SetKeyCodesAreScanCodes(TBool aEnable){iKeyCodesAreScanCodes = aEnable;}

	void ActivateL();
	void Deactivate();

	void SetVirtualButtons(const TVirtualButton* aVirtualButtons){iVirtualButtons = aVirtualButtons;}

	TBool				IsHotKeyForVirtualKeyboard(TInt aKeyCode){return (aKeyCode == iVkHotKeyKeyCode);}
	void				SetHotKeyForVirtualKeyboard(TInt aKeyCode){iVkHotKeyKeyCode = aKeyCode;}

public:			// implements MGameScreenDrawObserver
	virtual TBool	WantDrawToScreen();
	virtual void	PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice);

public:			// imlements MGameControlRawInputHandler
	virtual TBool ConsumeRawKeyEvent(TInt /*aType*/, const TKeyEvent& /*aKeyEvent*/){return EFalse;}
	virtual TBool ConsumeRawPointerEvent(const TPointerEvent& aPointerEvent);

public:			// imlements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
	virtual void HandlePointerEvent(const TPointerEvent& aPointerEvent);

public:		// implements MScreenOrientationObserver
	virtual void ScreenOrientationChanged(TOrientation aOrientation);

protected:
	CVirtualKeyboard(CGameWindow& aGameWindow, CGameController& aGameController);
	void ConstructL();

	TInt				KeyCode(const TVirtualKey* aVirtualKey);
	TInt				ScanCode(const TVirtualKey* aVirtualKey);
	void				CheckForVirtualKeyboardHotKey(TInt aKeyCode);
	void				HandleVirtualKeyDown(const TVirtualKey* aVirtualKey);
	void				HandleVirtualKeyUp();
	const TVirtualKey*	FindVirtualKey(const TPoint& aPoint);
	const TVirtualKey*	iKey;


	TBool		iCapsMode;
	TBool		iShiftMode;
	TBool		iSymbolMode;

	TBool		iKeyCodesAreScanCodes;

	void		DrawKeyboard();
	void		ClearKeyboard();


	TInt					KeyCode(const TVirtualButton* aVirtualButton);
	TInt					ScanCode(const TVirtualButton* aVirtualButton);
	TBool					CheckVirtualButton(const TPointerEvent& aPointerEvent);
	void					HandleVirtualButtonDown(const TVirtualButton* aVirtualButton);
	void					HandleVirtualButtonUp();
	const TVirtualButton*	FindVirtualButton(const TPoint& aPoint);
	const TVirtualButton*	iButton;


	CFbsGameScreenDevice*		iBitmap;
	CFont*						iFont;
	TRect						iRect;
	CGameWindow&				iGameWindow;
	CGameController&			iGameController;
	TBool						iIsActive;
	TRgb						iFgColor;
	TRgb						iBgColor;

	TInt						iAlpha;
	TInt						iAlphaInc;

	TRect						iActiveRect;

	TOrientation				iOrientation;

	const TVirtualButton*		iVirtualButtons;

	TInt						iVkHotKeyKeyCode;
};



#define MAX_SIMULTANEOUS_KEYPRESSES 5

class CMameStyleInputHandler : public CBase, public MGameControlInputHandler
{
public:
	CMameStyleInputHandler();
	TBool IsKeyPressed(TInt aKeyCode);
	TBool IsAnyKeyPressed();

public:		// implements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
	virtual void HandlePointerEvent(const TPointerEvent& /*aPointerEvent*/){};

protected:
	void	KeyDown(TInt aScanCode, TInt aKeyCode);
	void	KeyUp(TInt aScanCode);
	TBool	DelayKeyUpEvent(TInt aScanCode);
	void	ForceDelayedKeyUpEvents();

	TInt	iKeyCodes[MAX_SIMULTANEOUS_KEYPRESSES];
	TInt	iScanCodes[MAX_SIMULTANEOUS_KEYPRESSES];
	TInt	iLastScanCode;
	TInt	iNoOfKeyPresses;
	TInt	iNoIsKeyPressed;
};


class CSmartPenInputHandler : public CBase, 
		public MGameControlInputHandler, 
		public MGameScreenDrawObserver, 
		public MScreenOrientationObserver
{
public:
	~CSmartPenInputHandler();
	TBool IsPenDown(){return iPenDown;}

public:		// implements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt /*aType*/, const TKeyEvent& /*aKeyEvent*/){};
	virtual void HandlePointerEvent(const TPointerEvent& aPointerEvent);

public:		// implements MGameScreenDrawObserver
	virtual void PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice);

public:		// implements MScreenOrientationObserver
	virtual void ScreenOrientationChanged(TOrientation aOrientation){iOrientation=aOrientation;}

protected:
	virtual void HandleMoveEvent(TInt aX, TInt aY) = 0;
	virtual void HandleUpEvent() = 0;
	virtual void HandleDownEvent() = 0;

protected:
	CSmartPenInputHandler(CGameWindow& aGameWindow, CGameController& aGameController);
	void ConstructL();

	void SimulateAutoMove();
	void PenMoveEvent(const TPoint& aPosition, TBool aStrict = ETrue);
	void PenUpEvent(const TPoint& aPosition);
	void PenDownEvent(const TPoint& aPosition);
	TPoint TransformDeltaForOrientation(const TPoint& aPoint);

	TOrientation	iOrientation;

	TPoint		iLastPenPosition;
	TPoint		iDownPenPosition;
	TPoint		iLastDelta;
	TBool		iPenDown;
	TInt		iEventCount;

	TBool		iAutoX;
	TBool		iAutoY;

	TInt		iFramesThreshHold;
	TInt		iFramesRepeatThreshHold;
	TInt		iFramesMoveThreshHold;
	TInt		iFrameCount;

	CGameController&			iGameController;
	CGameWindow&				iGameWindow;
};


class CSmartPenToKeyInputHandler : public CSmartPenInputHandler
{
public:
	static CSmartPenToKeyInputHandler* NewL(CGameWindow& aGameWindow, CGameController& aGameController);
	void CanDo_X_And_Y_AtTheSameTime(TBool aCanDo){iCanDo_X_And_Y_AtTheSameTime = aCanDo;}
	void AutoForward(TBool aEnable) {iEnableAutoForward=aEnable;iCanDo_X_And_Y_AtTheSameTime=ETrue;}

protected:
	virtual void HandleMoveEvent(TInt aX, TInt aY);
	virtual void HandleUpEvent();
	virtual void HandleDownEvent();

	void HandleVertMoveEvent(TInt aNewVertMove);
	void HandleHorMoveEvent(TInt aNewHorMove);

protected:
	CSmartPenToKeyInputHandler(CGameWindow& aGameWindow, CGameController& aGameController);
	void ConstructL();

	void InjectkeyEvent(TInt aEventType, TInt aKeyOrScanKey, TInt aRepeat=0);
	TBool				iCanDo_X_And_Y_AtTheSameTime;

	TInt		iHorMove;
	TInt		iVertMove;

	TInt		iLeftCount;
	TInt		iRightCount;
	TInt		iUpCount;
	TInt		iDownCount;

	TBool		iIsMovingVert;
//	TBool		iFirstMove;
	TBool		iEnableAutoForward;
	TPoint		iLastPenPosition1;
};




/**********************************************
 *
 * CFpsMgr
 *
 **********************************************/

class CInfoMgr : public CBase, public MGameControlInputHandler, MGameControlRawInputHandler, MGameScreenDrawObserver
{
public:
	~CInfoMgr();
	static CInfoMgr* NewL(CGameWindow& aGameWindow, CGameController& aGameController);

public:			// implements MGameScreenDrawObserver
	virtual void PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice);

public:			// imlements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
	virtual void HandlePointerEvent(const TPointerEvent& /*aPointerEvent*/){};
	virtual TBool ConsumeRawKeyEvent(TInt /*aType*/, const TKeyEvent& /*aKeyEvent*/){ return EFalse;}
	virtual TBool ConsumeRawPointerEvent(const TPointerEvent& aPointerEvent){return EFalse;};

protected:
	CInfoMgr(CGameWindow& aGameWindow, CGameController& aGameController);
	void ConstructL();

	void ActivateL();
	void Deactivate();

	void DrawBitmap();
	void GameTick();

	TInt		iVideoFps;
	TInt		iPeriod;
	TInt		iPrevTicks;
	TInt		iFrames;

	CFbsGameScreenDevice*		iBitmap;
	CFont*						iFont;
	TPoint						iPosition;
	TSize						iSize;
	CGameWindow&				iGameWindow;
	CGameController&			iGameController;
	TBool						iIsActive;
	TRgb						iFgColor;
	TRgb						iBgColor;
};


/**********************************************
 *
 * CGameIntro
 *
 **********************************************/

class CGameIntro : public CBase, public MGameControlInputHandler, MGameScreenDrawObserver
{
public:
	~CGameIntro();
	static CGameIntro* NewL(CGameWindow& aGameWindow, CGameController& aGameController);

public:			// implements MGameScreenDrawObserver
	virtual TBool	WantDrawToScreen();
	virtual void	PostDrawFrameToScreen(CGameScreenDevice* aDstGameScreenDevice);

public:			// imlements MGameControlInputHandler
	virtual void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
	virtual void HandlePointerEvent(const TPointerEvent& /*aPointerEvent*/){};

protected:
	CGameIntro(CGameWindow& aGameWindow, CGameController& aGameController);
	void ConstructL();
	void Destruct();

	void ActivateL();
	void Deactivate();

	void DrawBitmap();
	void BitmapRecolor();

	CFbsGameScreenDevice*		iBitmap;
	CFont*						iFont;
	TPoint						iPosition;
	TSize						iSize;
	CGameWindow&				iGameWindow;
	CGameController&			iGameController;
	TBool						iIsActive;
	TRgb						iFgColor;
	TRgb						iBgColor;
	TInt						iFrames;
};


class CGameDirectScreenAccess : public CBase, public MGameScreenDrawObserver
{
public:
	~CGameDirectScreenAccess();
	static CGameDirectScreenAccess*	NewL(CGameWindow& aGameWindow);

public:			// MGameScreenDrawObserver
	virtual void PreDrawFrameToScreen(CGameScreenDevice* /*aDstGameScreenDevice*/);

protected:
	CGameDirectScreenAccess(CGameWindow& aGameWindow);
	void ConstructL();

	void ClaimScreenL();
	void PauseGameL();
	void ResumeGameL();

	RDirectScreenAccess		iDirectScreenAccess;
	RSemaphore				iSemaphore;
	TRequestStatus			iRequestStatus;
	CGameWindow&			iGameWindow;
};


#endif			/* __VIDEO_H */
